#  examples.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
import cervmongo

AsyncIODoc = cervmongo.get_async_doc()
SyncIODoc = cervmongo.get_doc()

def generate_custom_sync_doc(name:str, unique_id:str=None):
    class CustomDoc(SyncIODoc):
        """
        {} Document object.
        """.format(name.title())
        __name__ = name
        UNIQUE_ID = unique_id
        DEFAULT_COLLECTION = DC = "_".join(name.lower().strip().split()) # INFO: required. supply as class attr, or pass on to __init__ kwarg 'collection'
        TEMPLATE_PATH = {"created_datetime": None} # INFO: path to example json document or dict
        SETTINGS_COLLECTION = None # REC: when saving settings related to document type for WebApp
        SCHEMA_PATH = None # INFO: used when strict schema in place, otherwise just use TEMPLATE_PATH for rapid, agile development

        @router.get("/purchases/{_id}", tags=["purchases"], response_model=ResponseDocument)
        async def get_document(self, _id:str):
            """Return document by unique id, if supplied, else '_id' field"""
            return self.load(_id)

        @router.post("/purchases/{_id}", tags=["purchases"], response_model=ResponseDocument)
        async def create_document(self, _id:str = None, datetime:str=None, username:str=None, user_collection:str=None, user_unique_id_field:str=None, **kwargs):
            """Create simple new document"""

            if self.UNIQUE_ID: # INFO: assign _id to UNIQUE_ID if supplied, else '_id'
                kwargs[self.UNIQUE_ID] = _id
            else:
                kwargs["_id"] = _id

            self.create(**kwargs)

            if username:
                user_relation = {
                    "collection": user_collection,
                    "field": user_unique_id_field,
                    "value": username
                    }
            else:
                user_relation = {}

            self.add_timestamp("created_datetime", value=datetime, relation=user_relation)

            return self.save()

    return CustomDoc

def generate_custom_async_doc():
    pass



#  aio.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
__all__ = ["SUPPORT_ASYNCIO_CLIENT", "SUPPORT_ASYNCIO_BUCKET", "get_async_client", "get_async_doc", "AsyncIOClient", "AsyncIODoc"]

from os import path as os_path
from pymongo import WriteConcern
from dateutil.parser import parse as dateparse
import types
from jsonschema import validate
from functools import partial
import typing
import copy
import logging

from .models import (
                GenericResponse,
                StandardResponse,
                MongoListResponse,
                MongoDictResponse,
                )
from .vars import (
                StringEnum,
                IntEnum,
                PAGINATION_SORT_FIELDS,
                ENUM,
                DOC_ID,
                )
from .utils import (
                get_file_meta_information,
                parse_string_header,
                format_string_for_id,
                silent_drop_kwarg,
                current_datetime,
                file_and_fileobj,
                detect_mimetype,
                dict_to_query,
                clean_kwargs,
                current_date,
                json_load,
                json_dump,
                logger,
                )
from .config import Config

try:
    from motor.motor_asyncio import AsyncIOMotorClient as MongoClient
    from motor.motor_asyncio import AsyncIOMotorGridFSBucket as GridFSBucket
    SUPPORT_ASYNCIO_CLIENT = True #: True if motor package is installed else False
    SUPPORT_ASYNCIO_BUCKET = True #: True if motor package is installed else False
except:
    logger.warning("motor is not installed. needed if using asyncio")
    class MongoClient: pass
    class GridFSBucket: pass # NOTE: in case of refereneces
    SUPPORT_ASYNCIO_CLIENT = False #: True if motor package is installed else False
    SUPPORT_ASYNCIO_BUCKET = False #: True if motor package is installed else False


class AsyncIOClient(MongoClient):
    """
High-level AsyncIOMotorClient subclass with additional methods added for ease-of-use,
having some automated conveniences and defaults.
    """
    _MONGO_URI = lambda _: getattr(Config, "MONGO_URI", None)
    _DEFAULT_COLLECTION = None
    _KWARGS = None
    _LOGGING_COND_GET = None
    _LOGGING_COND_POST = None
    _LOGGING_COND_PUT = None
    _LOGGING_COND_PATCH = None
    _LOGGING_COND_DELETE = None

    def __init__(self, mongo_uri=None, default_collection=None, **kwargs):
        self._MONGO_URI = mongo_uri or self._MONGO_URI
        if callable(self._MONGO_URI):
            self._MONGO_URI = self._MONGO_URI()
        self._DEFAULT_COLLECTION = default_collection or self._DEFAULT_COLLECTION

        if kwargs:
            self._KWARGS = kwargs.copy()

        for kwarg in kwargs.keys():
            if kwarg.lower() in ('logging_cond_get', 'logging_cond_post',
                                'logging_cond_put', 'logging_cond_patch',
                                'logging_cond_delete'):
                setattr(self, kwarg.upper(), kwargs.pop(kwarg))

        MongoClient.__init__(self, self._MONGO_URI, **kwargs)

        db = self.get_default_database()
        logger.info("db detected '{}' of type '{}'".format(db.name, type(db.name)))
        if not getattr(db, "name", None) or db.name == "None":
            logger.warning("database not provided in MONGO_URI, assign with method set_database")
            logger.warning("gridfsbucket not instantiated due to missing database")
        else:
            global SUPPORT_ASYNCIO_BUCKET
            if SUPPORT_ASYNCIO_BUCKET:
                logger.debug("gridfsbucket instantiated under self.FILES")
                self.FILES = GridFSBucket(db)
            else:
                logger.warning("gridfsbucket not instantiated due to missing 'tornado' package")
                self.FILES = None

    def __repr__(self):
        db = self.get_default_database()

        if not getattr(db, "name", None) or db.name == "None":
            return "<cervmongo.AsyncIOClient>"
        else:
            return f"<cervmongo.AsyncIOClient.{db.name}>"

    def _process_record_id_type(self, record):
        one = False
        if isinstance(record, str):
            one = True
            if "$oid" in record:
                record = json_load(record)
            else:
                try:
                    record = DOC_ID.__supertype__(record)
                except:
                    pass
        elif isinstance(record, DOC_ID.__supertype__):
            record = record
            one = True
        elif isinstance(record, dict):
            if "$oid" in record or "$regex" in record:
                record = json_dump(record)
                record = json_load(record)
                one = True
        return (record, one)

    def set_database(self, database):
        Config.set_mongo_db(database)
        if self._KWARGS:
            AsyncIOClient.__init__(self, mongo_uri=Config.MONGO_URI, default_collection=self._DEFAULT_COLLECTION, **self._KWARGS)
        else:
            AsyncIOClient.__init__(self, mongo_uri=Config.MONGO_URI, default_collection=self._DEFAULT_COLLECTION)

    def COLLECTION(self, collection:str):

        self._DEFAULT_COLLECTION = collection


        class CollectionClient:
            __parent__ = CLIENT = self
            # INFO: variables
            _DEFAULT_COLLECTION = collection
            _MONGO_URI = self._MONGO_URI
            # INFO: general methods
            GENERATE_ID = self.GENERATE_ID
            COLLECTION = self.COLLECTION
            # INFO: GridFS file operations
            UPLOAD = self.UPLOAD
            DOWNLOAD = self.DOWNLOAD
            ERASE = self.ERASE
            # INFO: truncated Collection methods
            INDEX = partial(self.INDEX, collection)
            ADD_FIELD = partial(self.ADD_FIELD, collection)
            REMOVE_FIELD = partial(self.REMOVE_FIELD, collection)
            DELETE = partial(self.DELETE, collection)
            GET = partial(self.GET, collection)
            POST = partial(self.POST, collection)
            PUT = partial(self.PUT, collection)
            PATCH = partial(self.PATCH, collection)
            REPLACE = partial(self.REPLACE, collection)
            SEARCH = partial(self.SEARCH, collection)
            PAGINATED_QUERY = partial(self.PAGINATED_QUERY, collection)
            def __repr__(s):
                return "<cervmongo.AsyncIOClient.CollectionClient>"
            def get_client(s):
                return s.CLIENT
        return CollectionClient()

    async def PAGINATED_QUERY(self, collection, limit:int=20,
                                sort:PAGINATION_SORT_FIELDS=PAGINATION_SORT_FIELDS["_id"],
                                after:str=None, before:str=None,
                                page:int=None, endpoint:str="/",
                                ordering:int=-1, query:dict={}, **kwargs):
        """
            Returns paginated results of collection w/ query.

            Available pagination methods:
             - **Cursor-based (default)**
                - after
                - before
                - limit (results per page, default 20)
             - **Time-based** (a datetime field must be selected)
                - sort (set to datetime field)
                - after (records after this time)
                - before (records before this time)
                - limit (results per page, default 20)
             - **Offset-based** (not recommended)
                - limit (results per page, default 20)
                - page
        """
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        if isinstance(sort, ENUM.__supertype__):
            sort = sort.value

        total_docs = await self.GET(collection, query, count=True, empty=0)

        if not page:
            if sort == "_id":
                pagination_method = "cursor"
            else:
                pagination_method = "time"
            cursor = await self.GET(collection, query,
                                    limit=limit, key=sort, before=before,
                                    after=after, sort=ordering, empty=[])
        else:
            assert page >= 1, "page must be equal to or greater than 1"
            pagination_method = "offset"
            cursor = await self.GET(collection, query,
                                    perpage=limit, key=sort, page=page,
                                    sort=ordering, empty=[])

        results = [ record async for record in cursor ]

        # INFO: determine 'cursor' template
        if sort == "_id":
            template = "_{_id}"
        else:
            template = "{date}_{_id}"

        new_after = None
        new_before = None

        if results:
            _id = results[-1]['_id']
            try:
                date = results[-1][sort].isoformat()
            except:
                date = None
            if len(results) == limit:
                new_after = template.format(_id=_id, date=date)

            _id = results[0]['_id']
            try:
                date = results[0][sort].isoformat()
            except:
                date = None
            if any((after, before)):
                new_before = template.format(_id=_id, date=date)

            if pagination_method in ("cursor", "time"):
                if before:
                    check_ahead = await self.GET(collection, query,
                                            limit=limit, key=sort, before=new_before, empty=0, count=True)
                    if not check_ahead:
                        new_before = None
                elif after:
                    check_ahead = await self.GET(collection, query,
                                            limit=limit, key=sort, after=new_after, empty=0, count=True)
                    if not check_ahead:
                        new_after = None

        response = {
            "data": results,
            "details": {
                "pagination_method": pagination_method,
                "query": dict_to_query(query),
                "sort": sort,
                "unique_id": getattr(self, "_UNIQUE_ID", "_id"),
                "total": total_docs,
                "count": len(results),
                "limit": limit
                }
            }

        endpoint = endpoint

        # TODO: Refactor
        if pagination_method in ("cursor", "time"):
            response["details"]["cursors"] = {
                  "after": new_after,
                  "before": new_before
                }
            before_url_template = "{endpoint}?sort={sort}&limit={limit}&before={before}"
            after_url_template = "{endpoint}?sort={sort}&limit={limit}&after={after}"
        else: # INFO: pagination_method == "offset"
            response["details"]["cursors"] = {
                  "prev_page": page - 1 if page > 1 else None,
                  "next_page": page + 1 if (page * limit) <= total_docs else None
                }
            before_url_template = "{endpoint}?sort={sort}&limit={limit}&page={page}"
            after_url_template = "{endpoint}?sort={sort}&limit={limit}&page={page}"

        if new_before:
            response["details"]["previous"] = before_url_template.format(
                                                                    endpoint=endpoint,
                                                                    sort=sort,
                                                                    page=page,
                                                                    limit=limit,
                                                                    after=new_after,
                                                                    before=new_before)
        else:
            response["details"]["previous"] = None

        if new_after:
            response["details"]["next"] = after_url_template.format(
                                                                    endpoint=endpoint,
                                                                    sort=sort,
                                                                    page=page,
                                                                    limit=limit,
                                                                    after=new_after,
                                                                    before=new_before)
        else:
            response["details"]["next"] = None

        return response
    PAGINATED_QUERY.clean_kwargs = lambda kwargs: _clean_kwargs(ONLY=("limit", "sort", "after",
                                            "before", "page", "endpoint", "query"), kwargs=kwargs)

    def GENERATE_ID(self, _id=None):
        if _id:
            return DOC_ID.__supertype__(_id)
        else:
            return DOC_ID.__supertype__()

    async def UPLOAD(self, fileobj, filename:str=None, content_type:str=None, extension:str=None, **kwargs):
        assert self.FILES, "GridFS instance not initialized, run method 'set_database' with the desired database and try again"
        fileobj = file_and_fileobj(fileobj)
        metadata = get_file_meta_information(fileobj, filename=filename, content_type=content_type, extension=extension)
        filename = metadata['filename']
        metadata.update(kwargs)
        file_id = await self.FILES.upload_from_stream(filename, fileobj, metadata=metadata)
        return file_id

    async def ERASE(self, filename_or_id, revision:int=-1):
        assert self.FILES, "GridFS instance not initialized, run method 'set_database' with the desired database and try again"
        fs_doc = await self.DOWNLOAD(filename_or_id, revision=revision)
        await self.FILES.delete(fs_doc._id)
        await fs_doc.close()

    async def DOWNLOAD(self, filename_or_id=None, revision:int=-1, skip:int=None, limit:int=None, sort:int=-1, **query):
        assert self.FILES, "GridFS instance not initialized, run method 'set_database' with the desired database and try again"
        revision = int(revision)
        if filename_or_id:
            if isinstance(filename_or_id, DOC_ID.__supertype__):
                return await self.FILES.open_download_stream(filename_or_id)
            else:
                return await self.FILES.open_download_stream_by_name(filename_or_id, revision=revision)

        return self.FILES.find(query, limit=limit, skip=skip, sort=sort, no_cursor_timeout=True)

    async def DELETE(self, collection, record, soft:bool=False):
        db = self.get_default_database()
        if not collection:
            if hasattr(self, '_DEFAULT_COLLECTION'):
                collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        collection = db[collection]

        record = self._process_record_id_type(record)[0]

        if soft:
            data_record = await self.GET(collection, record)
            try:
                await self.PUT("deleted."+collection, data_record)
            except:
                data_record.pop("_id")
                await self.PUT("deleted."+collection, data_record)

        if isinstance(record, (str, ObjectId)):
            return await collection.delete_one({'_id': record})
        elif isinstance(record, dict):
            return await collection.delete_one(record)
        else:
            results = []
            for _id in record:
                results.append(await collection.delete_one({'_id': _id}))
            return results

    def INDEX(self, collection, key:str="_id", sort:int=1, unique:bool=False, reindex:bool=False):
        db = self.get_default_database()
        if not collection:
            if hasattr(self, '_DEFAULT_COLLECTION'):
                collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        collection = db[collection]

        name = "%sIndex%s" % (key, "Asc" if sort == 1 else "Desc")
        try:
            if not name in collection.index_information():
                collection.create_index([
                    (key, sort)], name=name, background=True, unique=unique)
        except:
            #print((_traceback()))
            pass

    async def ADD_FIELD(self, collection, field:str, value:typing.Union[typing.Dict, typing.List, str, int, float, bool]='', data=False, query:dict={}):
        if not collection:
            if hasattr(self, '_DEFAULT_COLLECTION'):
                collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        query.update({field: {'$exists': False}})
        if data:
            records = await self.GET(collection, query, fields={
                data: True}, empty=[])
        else:
            records = await self.GET(collection, query, fields={
                '_id': True}, empty=[])

        for record in records:
            if data:
                await self.PATCH(collection, record['_id'], {"$set": {
                    field: record[data]}})
            else:
                await self.PATCH(collection, record['_id'], {"$set": {
                    field: value}})

    async def REMOVE_FIELD(self, collection, field:str, query:dict={}) -> None:
        if not collection:
            collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        query.update({field: {'$exists': True}})
        records = await self.GET(collection, query, distinct=True)

        for record in records:
            await self.PATCH(collection, record, {"$unset": {field: ""}})

    async def GET(self, collection, id_or_query:typing.Union[DOC_ID, str, typing.Dict]={}, sort:int=1, key:str="_id", count:bool=None, search:str=None, fields:dict=None, page:int=None, perpage:int=False, limit:int=None, after:str=None, before:str=None, empty=None, distinct:str=None, one:bool=False, **kwargs):
        db = self.get_default_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection not provided"

        if not isinstance(collection, (list, tuple, types.GeneratorType)):
            collection = [collection]
        cols = list(set(collection))
        results = []
        number_of_results = len(cols)

        if distinct == True:
            distinct = "_id"

        id_or_query, _one = self._process_record_id_type(id_or_query)
        one = _one if _one else one
        if _one:
            query = {"_id": id_or_query}
        else:
            query = id_or_query

        for collection in cols:
            collection = db[collection]

            if query or not search:
                if count and not limit:
                    if query:
                        results.append(await collection.count_documents(query, **kwargs))
                    else:
                        results.append(await collection.estimated_document_count(**kwargs))
                elif distinct:
                    cursor = await collection.distinct(distinct, filter=query, **kwargs)
                    results.append(sorted(cursor))
                elif perpage:
                    total = (page - 1) * perpage
                    cursor = collection.find(query, fields, **kwargs)
                    results.append(cursor.sort([(key, sort)]).skip(total).limit(perpage))
                elif limit:
                    if any((query, after, before)):
                        query = {"$and": [
                                    query
                                ]}
                    if after or before:
                        if after:
                            sort_value, _id_value = after.split("_")
                            _id_value = DOC_ID.__supertype__(_id_value)
                            query["$and"].append({"$or": [
                                            {key: {"$lt": _id_value}}
                                        ]})
                            if key != "_id":
                                sort_value = dateparse(sort_value)
                                query["$and"][-1]["$or"].append({key: {"$lt": sort_value}, "_id": {"$lt": _id_value}})
                        elif before:
                            sort_value, _id_value = before.split("_")
                            _id_value = DOC_ID.__supertype__(_id_value)
                            query["$and"].append({"$or": [
                                            {key: {"$gt": _id_value}}
                                        ]})
                            if key != "_id":
                                sort_value = dateparse(sort_value)
                                query["$and"][-1]["$or"].append({key: {"$gt": sort_value}, "_id": {"$gt": _id_value}})

                    if count:
                        try:
                            cursor = await collection.count_documents(query, limit=limit, hint=[(key, sort)], **kwargs)
                        except:
                            cursor = len(await collection.find(query, fields, **kwargs).sort([(key, sort)]).to_list(limit))
                        results.append(cursor)
                    else:
                        cursor = collection.find(query, fields, **kwargs).sort([(key, sort)]).limit(limit)
                        results.append(cursor)
                elif one:
                    val = await collection.find_one(query, fields, **kwargs)
                    results.append(val if val else empty)
                else:
                    cursor = collection.find(query, fields, **kwargs).sort([(key, sort)])
                    results.append(cursor)
            elif search:
                try:
                    if count:
                        results.append(await cursor.count_documents({"$text": {"$search": search}}))
                    elif distinct:
                        results.append(await collection.distinct(distinct, filter={"$text": {"$search": search}}))
                    else:
                        cursor = collection.find({"$text": {"$search": search}})
                        if perpage:
                            total = (page - 1) * perpage
                            results.append(cursor.sort([(key, sort)]).skip(total).limit(perpage))
                        else:
                            results.append(cursor.sort([(key, sort)]))
                except:
                    cursor = await collection.command('textIndex', search=search)
                    if count:
                        results.append(cursor.count())
                    elif distinct:
                        results.append(cursor.distinct(distinct))
                    else:
                        if perpage:
                            total = (page - 1) * perpage
                            results.append(cursor.sort([(key, sort)]).skip(total).limit(perpage))
                        else:
                            results.append(cursor.sort([(key, sort)]))
            else:
                raise Error("unidentified error")

        if number_of_results == 1:
            return results[0]
        else:
            return results

    async def SEARCH(self, collection, search:str, **kwargs):
        if not collection:
            if hasattr(self, '_DEFAULT_COLLECTION'):
                collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        return await self.GET(collection, search=search, **kwargs)

    async def POST(self, collection, record_or_records:typing.Union[typing.List, typing.Dict]):
        db = self.get_default_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        if isinstance(record_or_records, (list, tuple)):
            return await collection.insert_many(record_or_records)
        elif isinstance(record_or_records, dict):
            return await collection.insert_one(record_or_records)
        else:
            raise TypeError("invalid record type '{}' provided".format(type(record_or_records)))

    async def PUT(self, collection, record_or_records:typing.Union[typing.List, typing.Dict]):
        db = self.get_default_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        if isinstance(record_or_records, (list, tuple)):
            return await collection.insert_many(record_or_records)
        elif isinstance(record_or_records, dict):
            return await collection.insert_one(record_or_records)
        else:
            raise TypeError("invalid record type '{}' provided".format(type(record_or_records)))

    async def REPLACE(self, collection, original, replacement:dict, upsert=False):
        db = self.get_default_database()
        if not collection:
            if hasattr(self, '_DEFAULT_COLLECTION'):
                collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        collection = db[collection]

        return await collection.replace_one({'_id': original},
                    replacement, upsert=upsert)

    async def PATCH(self, collection, original, updates, upsert:bool=False, w:int=1, multi:bool=False, log=False):
        db = self.get_default_database()
        if not collection:
            if hasattr(self, '_DEFAULT_COLLECTION'):
                collection = self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        collection = db[collection]

        if w != 1:
            WRITE = WriteConcern(w=0)
            _collection = collection.with_options(write_concern=WRITE)
        else:
            _collection = collection
        if log:
            try:
                pass # TODO: optional, may not incorporate
            except:
                pass

        original = self._process_record_id_type(original)

        if multi:
            results = await _collection.update_many(original, updates, upsert=upsert)

            return results
        elif isinstance(original, (str, ObjectId)):
            query = {'_id': original}
            updates["$setOnInsert"] = query
            results = await _collection.update_one(query, updates, upsert=upsert)
            return results
        else:
            results = []
            for i, _id in enumerate(original):
                _id = self._process_record_id_type(_id)
                query = {'_id': _id}
                updates["$setOnInsert"] = query
                result = await collection.update_one(query, updates[i], upsert=upsert)
                results.append(result)
            await self.logs.other.insert_one({'name': 'reindex',
                                        'db': db,
                                        'collection': collection,
                                        'datetime': current_datetime()})
            return results



class AsyncIODoc(AsyncIOClient):
    """
        Custom MongoClient subclass with customizations for creating
        standardized documents and adding json schema validation.
    """
    _MONGO_URI = lambda _: getattr(Config, "MONGO_URI", None)
    _DOC_TYPE:str = None #: MongoDB collection to use
    _DOC_ID:str = "_id"
    _DOC_SAMPLE:str = None
    _DOC_SCHEMA:str = None
    _DOC_MARSHMALLOW:str = False
    _DOC_DEFAULTS:dict = {}
    _DOC_RESTRICTED_KEYS:list = []
    _DOC_ENUMS:list = []
    _DOC_SETTINGS:str = None

    def __init__(self, _id=None, doc_type:str=None, doc_sample:typing.Union[typing.Dict, str]=None, doc_schema:typing.Union[typing.Dict, str]=None, doc_id:str=None, mongo_uri:str=None, **kwargs):
        self._MONGO_URI = mongo_uri or self._MONGO_URI
        if callable(self._MONGO_URI):
            self._MONGO_URI = self._MONGO_URI()
        # INFO: set default collection
        self._DOC_TYPE = doc_type or self._DOC_TYPE
        assert self._DOC_TYPE, "collection must be of type str"
        self._DEFAULT_COLLECTION = self._DOC_TYPE
        # INFO: location for sample record, used as template
        self._DOC_SAMPLE = doc_sample or self._DOC_SAMPLE
        # INFO: path to validation schema
        self._DOC_SCHEMA = doc_schema or self._DOC_SCHEMA
        # INFO: sets the unique id field for the document type, if any (cannot be _id)
        self._DOC_ID = doc_id or self._DOC_ID
        assert self._DOC_ID, "unique id field name must be of type str"

        for kwarg in kwargs.keys():
            if kwarg.lower() in ('doc_settings', 'doc_marshmallow', 'doc_defaults', 'doc_restricted_keys'):
                setattr(self, "_{}".format(kwarg.upper()), kwargs.pop(kwarg))

        # Initial Record object with sample else start blank dict
        if self._DOC_SAMPLE:
            if isinstance(self._DOC_SAMPLE, str):
                sample_full_path = os_path.join(Config.JSON_SAMPLE_PATH, self._DOC_SAMPLE)
                with open(sample_full_path) as file1:
                    sample = json_load(file1.read())
            elif isinstance(self._DOC_SAMPLE, dict):
                sample = self._DOC_SAMPLE
            else:
                raise TypeError("_DOC_SAMPLE is invalid type '{}', valid types are dict and str".format(type(self._DOC_SAMPLE)))
            sample_parent_found = sample.pop("__parent__", None)
            while sample_parent_found:
                parent_sample_full_path = os_path.join(Config.JSON_SAMPLE_PATH, sample_parent_found)
                with open(parent_sample_full_path) as _file:
                    parent_sample = json_load(_file.read())
                parent_sample.update(sample)
                sample = parent_sample
                sample_parent_found = sample.pop("__parent__", None)
            self.sample = sample
        else:
            self.sample = {}

        # INFO: Load schema else start blank dict to add manual validation entries
        if self._DOC_SCHEMA:
            if isinstance(self._DOC_SCHEMA, str):
                schema_full_path = os_path.join(Config.JSON_SCHEMA_PATH, self._DOC_SCHEMA)
                with open(schema_full_path) as _file:
                    self.schema = json_load(_file.read())
            elif isinstance(self._DOC_SCHEMA, dict):
                self.schema = self._DOC_SCHEMA
            else:
                raise TypeError("_DOC_SCHEMA is invalid type '{}', valid types are dict and str".format(type(self._DOC_SCHEMA)))
        else:
            self.schema = {}

        AsyncIOClient.__init__(self, **kwargs)

        db = self.get_default_database()

        if not getattr(db, "name", None) or db.name == "None":
            raise Exception("database not provided in MongoDB URI")
        else:
            self._DOC_DB = db.name

        # Initialize enums
        if not self._DOC_ENUMS:
            pass
            #enums_record = self.GET(self._DOC_SETTINGS, "enums"

        # INFO: If class has a _DOC_ID assigned, create unique index
        if self._DOC_ID != "_id":
            self.INDEX(self._DOC_TYPE, key=self._DOC_ID,
                                                        sort=1, unique=True)

        self.load(_id)

    def __repr__(self):
        if self.RECORD.get("_id", None):
            _id = self.id()
            return f"<cervmongo.AsyncIODoc.{self._DOC_DB}.{self._DOC_TYPE}.{self._DOC_ID}:{_id}>"
        else:
            return "<cervmongo.AsyncIODoc>"

    def __enter__(self):
        return self

    def __exit__(self):
        self.close()

    def _process_restrictions(self, record:dict=None):
        """removes restricted keys from record and return record"""
        try:
            if record:
                assert isinstance(record, (MongoDictResponse, dict)), "Needs to be a dictionary, got {}".format(type(record))
                return {key: value for key, value in record.items() if not key in self._DOC_RESTRICTED_KEYS}
            else:
                return {key: value for key, value in self.RECORD.items() if not key in self._DOC_RESTRICTED_KEYS}
        except:
            logger.exception("encountered error when cleaning self.RECORD, returning empty dict")
            return {}

    def _p_r(self, record:dict=None):
        """truncated alias for _process_restrictions"""
        return self._process_restrictions(record=record)

    def _generate_unique_id(self, template:str="{total}", **kwargs):
        return template.format(**kwargs).upper()

    def _timestamp(self, value=None):
        if value:
            try:
                value = dateparse(value)
            except:
                value = current_datetime()
        else:
            value = current_datetime()
        return value

    def _guess_corresponding_fieldname(self, _type="unknown", related_field:str=""):
        time_fields = ("date", "datetime", "time")
        if _type in time_fields:
            # NOTE: a timestamp is 'mostly' accompanied by a user or relation
            if related_field:
                if "_" in related_field:
                    field_parts = related_field.split("_")
                elif "-" in related_field:
                    field_parts = related_field.split("-")
                else:
                    field_parts = related_field.split()

                for field_part in field_parts:
                    field_part = field_part.strip(" _-").lower()
                    if any(x in field_part for x in time_fields):
                        continue
                    else:
                        return f"{field_part}_by"
                return "for"
            else:
                return "by"
        else:
            # NOTE: an unknown type field has a timestamp pairing or desc
            if related_field:
                return f"{related_field}_description"
            else:
                return "field_description"

    async def _related_record(self, collection:str=None, field:str="_id", value=False, additional:dict={}):
        additional.update({
            field: True
            })

        record = await self.GET(collection, {field: value}, fields=additional, one=True, empty={})
        assert record, 'Error: No record found'

        record['key'] = field
        record['collection'] = collection
        record[field] = record[field]

        return record

    async def load(self, _id=None):
        # If _id specified on init, load actual record versus blank template
        if _id:
            if self._DOC_ID:
                self.RECORD = await self.GET(self._DOC_TYPE, {self._DOC_ID: _id}, one=True)
            else:
                self.RECORD = await self.GET(self._DOC_TYPE, _id)
        else:
            self.RECORD = copy.deepcopy(self.sample)

        if not self.RECORD:
            self.RECORD = {}

        return StandardResponse(
                    data=self._p_r(self.RECORD),
                    details={
                "state": "unsaved" if not self.RECORD.get("_id", None) else "saved",
                "unique_id": self._DOC_ID
                })

    async def view(self, _id=False):
        if not _id:
            return StandardResponse(data=self._p_r(self.RECORD))
        else:
            if self._DOC_ID:
                return StandardResponse(
                            data=self._p_r(self.GET(self._DOC_TYPE, {self._DOC_ID: _id}, one=True, empty={})),
                            details={
                        "unique_id": self._DOC_ID
                        }
                    )
            else:
                return StandardResponse(
                            data=self._p_r(self.GET(self._DOC_TYPE, {"_id": _id}, one=True, empty={})),
                            details={
                        "unique_id": self._DOC_ID
                        }
                    )

    async def reload(self):
        assert self.RECORD.get('_id')
        self.RECORD = await self.GET(self._DOC_TYPE, self.RECORD['_id'])

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "unique_id": self._DOC_ID
                }
            }

    def id(self):
        return self.RECORD.get(self._DOC_ID, None)

    async def create(self, save:bool=False, trigger=None, template:str="{total}", query:dict={}, **kwargs):
        assert self.RECORD.get('_id') is None, """Cannot use create method on
 an existing record. Use patch method instead."""

        if self._DOC_MARSHMALLOW:
            self._DOC_MARSHMALLOW().load(kwargs)
            self.RECORD.update(kwargs)
        elif self.sample:
            # INFO: removing invalid keys based on sample record
            [ silent_drop_kwarg(kwargs, x, reason="not in self.sample") for x in list(kwargs.keys()) if not x in self.sample ]
            self.RECORD.update(kwargs)
        else:
            self.RECORD.update(kwargs)

        kwargs['total'] = str(self.GET(self._DOC_TYPE, query).count() + 1).zfill(6)

        if self._DOC_ID and not self.RECORD.get(self._DOC_ID):
            self.RECORD[self._DOC_ID] = self._generate_unique_id(template=template, **kwargs)

        if save:
            await self.save(trigger=None)

        return {
            "data": self._p_r(self.RECORD),
            "details": {"unique_id": self._DOC_ID, "collection": self._DOC_TYPE, "_id": self.RECORD[self._DOC_ID]}
            }

    async def add_enum(name:str, value):
        self.RECORD

    async def push(self, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use push method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        await self.PATCH(None, self.RECORD['_id'], {"$push": kwargs})
        await self.reload()

        keys = list(kwargs.keys())
        values = [ kwargs[key] for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "push",
                "desc": "push a value to end of array, not set",
                "field": kwargs.keys(),
                "value": kwargs.values()
                }
            }

    async def pull(self, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use pull method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        await self.PATCH(None, self.RECORD['_id'], {"$pull": kwargs})
        await self.reload()

        keys = list(kwargs.keys())
        values = [ kwargs[key] for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "pull",
                "desc": "pull all instances of a value from an array",
                "field": kwargs.keys(),
                "value": kwargs.values()
                }
            }

    async def increment(self, query:dict={}, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use increment method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        query.update({"_id": self.RECORD['_id']})

        await self.PATCH(None, query, {"$inc": kwargs}, multi=True)
        await self.reload()

        keys = list(kwargs.keys())
        values = [ kwargs[key] for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "increment",
                "desc": "increment the integer fields by the amount provided",
                "field": keys,
                "increment": values
                }
            }

    async def update(self, query:dict={}, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use update method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        query.update({"_id": self.RECORD['_id']})

        keys = list(kwargs.keys())
        old_values = [ self.RECORD.get(key, None) for key in keys ]

        await self.PATCH(None, query, {"$set": kwargs}, multi=True)
        await self.reload()

        new_values = [ self.RECORD.get(key, None) for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "update",
                "desc": "replace existing field values with new values",
                "field": keys,
                "values_old": old_values,
                "values_new": new_values
                }
            }

    async def patch(self, save:bool=False, trigger=None, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use patch method on
 a non-existing record.Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        if self._DOC_MARSHMALLOW:
            _DOC_MARSHMALLOW().load(kwargs, partial=True)
            self.RECORD.update(kwargs)
        elif self.sample:
            assert all([ x in self.sample for x in kwargs.keys()])
            self.RECORD.update(kwargs)
        else:
            self.RECORD.update(kwargs)
        if save:
            await self.save(trigger=trigger)

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "processed": True,
                "diff": kwargs
                }
            }

    async def save(self, trigger=None):
        _id = None

        if self._DOC_DEFAULTS:
            for key, value in self._DOC_DEFAULTS.items():
                if not self.RECORD.get(key):
                    self.RECORD[key] = value

        if self.RECORD.get('_id'):
            _id = self.RECORD.pop("_id")
        try:
            if self._DOC_MARSHMALLOW:
                self._DOC_MARSHMALLOW().load(self.RECORD)
            else:
                validate(self.RECORD, self.schema)
        except:
            raise
        else:
            if _id:
                await self.PATCH(None, _id, {"$set": self.RECORD})
                self.RECORD['_id'] = _id
            else:
                result = await self.PUT(None, self.RECORD)
                self.RECORD['_id'] = result.inserted_id
            if trigger:
                trigger()

        return {
            "data": self._p_r(self.RECORD),
            "details": {}
            }

    async def close(self):
        # TODO: clean closing logic
        await self.load()

        return {
            "data": self.RECORD,
            "details": {}
            }

    async def add_relation(self, key:str, *args, **kwargs):
        self.RECORD[key] = await self._related_record(*args, **kwargs)

        return {
            "data": self._p_r(self.RECORD),
            "details": {"key": key, "relation": self.RECORD[key]}
            }

    async def add_timestamp(self, key:str, value=None, relation:dict={}):
        self.RECORD[key] = self._timestamp(value)

        if relation:
            related_key = relation.pop("key", self._guess_corresponding_fieldname(_type="datetime", related_field=key))
            self.add_relation(related_key, **relation)
        return {
            "data": self._p_r(self.RECORD),
            "details": {"key": key, "value": self.RECORD[key]}
            }

    async def add_object(self, field:str, object_name:str=None, key:str=None, value=None, **kwargs):
        if not object_name:
            if not key:
                self.RECORD[field] = {}
            else:
                self.RECORD[field] = {key: value if value else self.GENERATE_ID()}
        else:
            template_path = os_path.join(_DOC_SAMPLE, object_name + ".json")
            assert os.path.exists(template_path), "path does not exist"
            with open(template_path) as file1:
                self.RECORD[field] = json_util.loads(file1.read())
            if key:
                self.RECORD[field][key] = value if value else self.GENERATE_ID()

        return {
            "data": self._p_r(self.RECORD),
            "details": {field: self.RECORD[field], "state": "unsaved"}
            }


def get_async_client() -> AsyncIOClient:
    """returns AsyncIOClient class"""
    global SUPPORT_ASYNCIO_CLIENT
    if not SUPPORT_ASYNCIO_CLIENT:
        raise Exception("motor not installed")
    return AsyncIOClient

def get_async_doc() -> AsyncIODoc:
    """returns AsyncIODoc class"""
    global SUPPORT_ASYNCIO_CLIENT
    if not SUPPORT_ASYNCIO_CLIENT:
        raise Exception("motor not installed")
    return AsyncIODoc



#  config.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
__all__ = ["Config"]

import os
from .utils import (
                    getenv_boolean,
                    logger,
                    json_dump,
                    )
from .vars import (
                    MONGODB_URI,
                    )
from .models import MetaConfig
import typing
import logging
ch = logging.StreamHandler()

ConfigClass = typing.TypeVar('Config')


class Defaults:
    MONGO_DB:typing.Optional[str] = os.getenv("MONGO_DB", None)
    MONGO_HOST:str = os.getenv("MONGO_HOST", "127.0.0.1")
    MONGO_PORT:int = int(os.getenv("MONGO_PORT", 27017))
    MONGO_REPLICA_SET:typing.Optional[str] = os.getenv("MONGO_REPLICA_SET", None)
    MONGO_MAX_POOL_SIZE:int = int(os.getenv("MONGO_MAX_POOL_SIZE", 20))
    MONGO_MIN_POOL_SIZE:int = int(os.getenv("MONGO_MIN_POOL_SIZE", 10))
    MONGO_USER:typing.Optional[str] = os.getenv("MONGO_USER", None)
    MONGO_PASSWORD:typing.Optional[str] = os.getenv("MONGO_PASSWORD", None)
    MONGO_URI:typing.Optional[str] = os.getenv("MONGO_URI", None)
    DEBUG_LEVEL:int = int(os.getenv("DEBUG_LEVEL", logging.WARNING))
    JSON_SAMPLE_PATH:str = os.getenv("JSON_SAMPLE_PATH", "./")
    JSON_SCHEMA_PATH:str = os.getenv("JSON_SCHEMA_PATH", "./")

class Config(metaclass=MetaConfig):
    """
        MongoDB and cervmongo settings, loaded initially by environmental variables
    """
    MONGO_DB:typing.Optional[str] = Defaults.MONGO_DB #: The MongoDB database to use
    MONGO_HOST:str = Defaults.MONGO_HOST #: The host server for MongoDB
    MONGO_PORT:int = Defaults.MONGO_PORT #: The port for connection to host server
    MONGO_REPLICA_SET:typing.Optional[str] = Defaults.MONGO_REPLICA_SET #: The name of the replica set, if any
    MONGO_MAX_POOL_SIZE:int = Defaults.MONGO_MAX_POOL_SIZE
    MONGO_MIN_POOL_SIZE:int = Defaults.MONGO_MIN_POOL_SIZE
    MONGO_USER:typing.Optional[str] = Defaults.MONGO_USER #: The username for MongoDB connection
    MONGO_PASSWORD:typing.Optional[str] = Defaults.MONGO_PASSWORD #: The password for MongoDB connection
    MONGO_URI:typing.Optional[str] = Defaults.MONGO_URI #: The MongoDB URI that will be used when accessing clients
    DEBUG_LEVEL:int = Defaults.DEBUG_LEVEL #: The level at which to display information, defaults to logging.warning
    JSON_SAMPLE_PATH:str = Defaults.JSON_SAMPLE_PATH #: For use with JSON sample records to simplify schema process of new documents
    JSON_SCHEMA_PATH:str = Defaults.JSON_SCHEMA_PATH #: For use with JSON schema documents for validating new documents on creation

    @classmethod
    def reset(cls) -> None:
        """
            resets config values to the first values assigned when cervmongo was imported
        """
        for attr, value in Defaults.__dict__.items():
            if attr.isupper():
                setattr(cls, attr, value)

    @classmethod
    def set_debug_level(cls, debug_level:int) -> None:
        """
            sets debug level of application, default is logging.warning
        """
        assert isinstance(debug_level, int), "debug level must be a valid logging int"
        cls.DEBUG_LEVEL = debug_level
        logger.setLevel(debug_level)
        ch.setLevel(debug_level)

    @classmethod
    def set_mongo_db(cls:typing.Type[ConfigClass], database_name:str) -> ConfigClass:
        """
            assigns the MONGO_DB var in config and regenerates mongodb uri

            returns Class
        """
        assert isinstance(database_name, str), "database must be a string"
        cls.MONGO_DB = database_name
        cls.generate_mongo_uri()
        return cls

    @classmethod
    def set_mongo_replica_set(cls:typing.Type[ConfigClass], replica_set=None) -> ConfigClass:
        """
            assigns the MONGO_REPLICA_SET var in config and regenerates mongodb uri

            returns Class
        """
        assert isinstance(replica_set, (str, type(None))), "database must be a string or NoneType"
        cls.MONGO_REPLICA_SET = replica_set
        cls.generate_mongo_uri()
        return cls

    @classmethod
    def generate_mongo_uri(cls) -> MONGODB_URI:
        """
            uses config class values to generate mongodb uri and returns connection string
        """
        if cls.MONGO_REPLICA_SET:
            cls.MONGO_URI = f"mongodb://{cls.MONGO_HOST}:{cls.MONGO_PORT}/{cls.MONGO_DB}?replicaSet={cls.MONGO_REPLICA_SET}"
        else:
            cls.MONGO_URI = f"mongodb://{cls.MONGO_HOST}:{cls.MONGO_PORT}/{cls.MONGO_DB}"
        return cls.MONGO_URI

    @classmethod
    def reload(cls) -> None:
        """
            Reloads config class values from environment, falls back to current values if none found
        """
        cls.MONGO_DB = os.getenv("MONGO_DB", cls.MONGO_DB)
        cls.MONGO_HOST = os.getenv("MONGO_HOST", cls.MONGO_HOST)
        cls.MONGO_PORT = int(os.getenv("MONGO_PORT", cls.MONGO_PORT))
        cls.MONGO_REPLICA_SET = os.getenv("MONGO_REPLICA_SET", cls.MONGO_REPLICA_SET)
        cls.MONGO_MAX_POOL_SIZE = int(os.getenv("MONGO_MAX_POOL_SIZE", cls.MONGO_MAX_POOL_SIZE))
        cls.MONGO_MIN_POOL_SIZE = int(os.getenv("MONGO_MIN_POOL_SIZE", cls.MONGO_MIN_POOL_SIZE))
        cls.MONGO_USER = os.getenv("MONGO_USER", cls.MONGO_USER)
        cls.MONGO_PASSWORD = os.getenv("MONGO_PASSWORD", cls.MONGO_PASSWORD)
        cls.MONGO_URI = os.getenv("MONGO_URI", cls.MONGO_URI)
        cls.DEBUG_LEVEL = int(os.getenv("DEBUG_LEVEL", cls.DEBUG_LEVEL))
        cls.JSON_SAMPLE_PATH = os.getenv("JSON_SAMPLE_PATH", cls.JSON_SAMPLE_PATH)
        cls.JSON_SCHEMA_PATH = os.getenv("JSON_SCHEMA_PATH", cls.JSON_SCHEMA_PATH)

    @classmethod
    def reload_from_file(cls, env_path:str=".env", override:bool=False) -> None:
        """
            Re-assign class variables using .env config file
        """
        import dotenv
        dotenv.load_dotenv(dotenv_path=env_path, override=override)
        cls.reload()

    @classmethod
    def reload_from_stream(cls, stream:typing.IO, override:bool=False) -> None:
        """
            Re-assign class variables using stream in .env config format
        """
        import dotenv
        dotenv.load_dotenv(stream=stream, override=override)
        cls.reload()

Config.generate_mongo_uri()

logger.setLevel(Config.DEBUG_LEVEL)
# create console handler and set level to debug
ch.setLevel(Config.DEBUG_LEVEL)
# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# add formatter to ch
ch.setFormatter(formatter)
# add ch to logger
logger.addHandler(ch)



#  models.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
import typing
from functools import partial
from .utils import snake2camel
from .utils import (
                    json_dump,
                    json_load,
                    yaml_dump,
                    yaml_load,
                    sort_list,
                    )
from .vars import (
                    DOC_ID,
                    DETAILS,
                    PAGINATION_SORT_FIELDS,
                    YAML,
                    JSON,
                    ASCENDING,
                    DESCENDING,
                    ObjectIdStr,
                    )
from pymongo.cursor import Cursor
try:
    from pydantic import BaseConfig, BaseModel
    SUPPORT_PYDANTIC = True

    class DefaultModel(BaseModel):
        """
        Intended for use as a base class for externally-facing models.
        Any models that inherit from this class will:
        * accept fields using snake_case or camelCase keys
        * use camelCase keys in the generated OpenAPI spec (when supported)
        * have orm_mode on by default

        - from fastapi-utils
        """

        class Config(BaseConfig):
            orm_mode = True
            allow_population_by_field_name = True
            alias_generator = partial(snake2camel, start_lower=True)
except:
    from dataclasses import dataclass
    SUPPORT_PYDANTIC = False

    @dataclass
    class DefaultModel:
        """
        Intended for use as a base class for externally-facing models.
        Any models that inherit from this class will:
        * accept fields using snake_case or camelCase keys
        * use camelCase keys in the generated OpenAPI spec (when supported)
        * create an orm_mode Config attrib and have it on by default

        - from fastapi-utils
        """

        class Config():
            orm_mode = True
            allow_population_by_field_name = True
            alias_generator = partial(snake2camel, start_lower=True)


# INFO: required to desired web response documents
class MetaConfig(type):
    def __str__(cls):
        return json_dump({
                        attr: value for attr, value in cls
                    }, pretty=True)

    def __iter__(cls):
        for attr, value in cls.__dict__.items():
            if attr.isupper():
                yield attr, value


class _GenericResponse(DefaultModel):
    def __repr__(self):
        return "GenericResponse"

    class Config:
        schema_extra = {
            'example': [
                {}
            ]
        }


class GenericResponse(_GenericResponse):
    """a premade web API friendly response object"""
    @property
    def __name__(self):
        return "GenericResponse"
    def __init__(self, *args, **kwargs):
        if SUPPORT_PYDANTIC:
            if args:
                data = args[0]
                if isinstance(data, (dict, _GenericResponse)):
                    return super().__init__(**dict(data))
                else:
                    return super().__init__(*data)
            else:
                return super().__init__(*args, **kwargs)
        else:
            super().__init__()
            if args:
                data = args[0]
                if isinstance(data, (dict, _GenericResponse)):
                    for data_kw in data:
                        setattr(self, data_kw, data[data_kw])
                        self.__dict__[data_kw] = data[data_kw]
            else:
                for kwarg in kwargs:
                    setattr(self, kwarg, kwargs[kwarg])
                    self.__dict__[kwarg] = kwargs[kwarg]


class _StandardResponse(DefaultModel):
    data: typing.Union[typing.Dict, typing.List, typing.Text, int, float, bool]
    details: DETAILS = DETAILS(...)

    def __repr__(self):
        return "StandardResponse"

    class Config:
        schema_extra = {
            'example': [
                {
                    'data': {"_id": "000000", "other": "data"},
                    'details': {"field": "other", "value": "data", "count": 1}
                }
            ]
        }
        json_encoders = {type(DOC_ID.__supertype__): lambda x: str(x)}
        orm_mode = True
        allow_population_by_field_name = True
        alias_generator = partial(snake2camel, start_lower=True)

class StandardResponse(_StandardResponse):
    """a premade web API friendly response object"""
    def __init__(self, *args, **kwargs):
        if SUPPORT_PYDANTIC:
            if args:
                data = args[0]
                if isinstance(data, (dict, _StandardResponse)):
                    return super().__init__(**dict(data))
                else:
                    return super().__init__(*data)
            else:
                return super(_StandardResponse, self).__init__(*args, **kwargs)
        else:
            super().__init__()
            if args:
                data = args[0]
                if isinstance(data, (dict, _StandardResponse)):
                    for data_kw in data:
                        setattr(self, data_kw, data[data_kw])
                        self.__dict__[data_kw] = data[data_kw]
            else:
                for kwarg in kwargs:
                    setattr(self, kwarg, kwargs[kwarg])
                    self.__dict__[kwarg] = kwargs[kwarg]


class YAMLStandardResponse(_StandardResponse):
    """a premade web API friendly response object"""
    def __init__(self, *args, **kwargs):
        if SUPPORT_PYDANTIC:
            if args:
                data = args[0]
                if isinstance(data, (dict, _StandardResponse)):
                    return super().__init__(**dict(data))
                else:
                    return super().__init__(*data)
            else:
                return super(_StandardResponse, self).__init__(*args, **kwargs)
        else:
            super().__init__()
            if args:
                data = args[0]
                if isinstance(data, (dict, _StandardResponse)):
                    for data_kw in data:
                        setattr(self, data_kw, data[data_kw])
                        self.__dict__[data_kw] = data[data_kw]
            else:
                for kwarg in kwargs:
                    setattr(self, kwarg, kwargs[kwarg])
                    self.__dict__[kwarg] = kwargs[kwarg]

    def __str__(self) -> YAML:
        return yaml_dump(self.dict())


class MongoDictResponse(dict):
    """the normal response for a single document when using cervmongo.main.SyncIOClient or cervmongo.aio.AsyncIOClient"""

    pass

class MongoListResponse(list):
    """the normal response for multiple documents when using cervmongo.main.SyncIOClient or cervmongo.aio.AsyncIOClient"""
    _sort = ASCENDING
    _index = 0
    def __init__(self, cursor=[]):
        if isinstance(cursor, Cursor):
            self._cursor = cursor
            self._cursor.rewind()
            self.__self__ = self._cursor
        else:
            self._cursor = None
            super().__init__(cursor)

    def __repr__(self):
        if self._cursor:
            return str(self._cursor)
        else:
            return super().__repr__()

    def __del__(self):
        if self._cursor:
            self._cursor.close()
        else:
            self.clear()

    def get(self) -> typing.Union[Cursor, typing.List]:
        """returns cursor or list instance"""
        if self._cursor:
            return self._cursor
        else:
            return self

    def close(self) -> None:
        """closes cursor or clears list"""
        if self._cursor:
            self._cursor.close()
        else:
            self.clear()

    # ~ @property
    # ~ def retrieved(self):
        # ~ """The number of documents retrieved so far"""
        # ~ if self._cursor:
            # ~ return self._cursor.retrieved
        # ~ else:
            # ~ return self._index + 1

    def rewind(self) -> None:
        """rewinds cursor, if any"""
        if self._cursor:
            self._cursor.rewind()

    def distinct(self, field:str="_id") -> typing.List[typing.Any]:
        """returns list of distinct values based on field, defaults to '_id'. supports dot notation for nested values."""
        if self._cursor:
            self._cursor.rewind()
            return sorted(self._cursor.distinct(field), reverse=True if self._sort == -1 else False)
        else:
            try:
                if "." in field:
                    results = set()
                    fields = field.split(".")
                    total = len(fields) - 1 # INFO: 0 start index
                    for item in self:
                        for index, field in enumerate(fields):
                            if not index == total:
                                if field.isdigit():
                                    try:
                                        item = item[int(field)]
                                    except:
                                        continue
                                elif field in item:
                                    item = item[field]
                                else:
                                    continue
                            else:
                                if field.isdigit():
                                    try:
                                        results.add(item.index(int(field)))
                                    except:
                                        continue
                                elif field in item:
                                    results.add(item[field])
                                else:
                                    continue
                    return sorted(list(results), reverse=True if self._sort == -1 else False)
                else:
                    return sorted([item[field] for item in self if field in item], reverse=True if self._sort == -1 else False)
            except:
                return []

    def count(self) -> int:
        """returns the count of the number of records in cursor or list results"""
        if self._cursor:
            try:
                self._cursor.rewind()
                return self._cursor.count(with_limit_and_skip=True)
            except:
                col = self._cursor._Cursor__collection
                query = self._cursor._Cursor__spec or {}
                limit = self._cursor._Cursor__limit
                sort = self._cursor._Cursor__ordering
                if sort:
                    sort = sort.items()
                return col.count_documents(query, limit=limit, hint=sort)
        else:
            return len(self)

    def sort(self, sort:int=1, key:str="_id") -> 'self':
        """sorts cursor results or list results, default is cervmongo.ASC (int 1)

        Options:
            - cervmongo.ASC or 1
            - cervmongo.DESC or -1
        """
        assert sort in (-1, 1), "sort must be option -1, 1"
        self._sort = sort
        if self._cursor:
            self._cursor.rewind()
            self._cursor = self._cursor.sort([(key, sort)])
            self.__self__ = self._cursor
            return self
        else:
            if sort == -1:
                super().sort(reverse=True, key=lambda item, field=key: sort_list(item, field))
            else:
                super().sort(key=lambda item, field=key: sort_list(item, field))
            return self

    def list(self) -> typing.List[typing.Dict]:
        """returns a new list representation of the current cursor"""
        if self._cursor:
            self._cursor.rewind()
            return MongoListResponse(list(self._cursor))
        else:
            return self





#  datatables.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#

# translation for sorting between datatables and mongodb
order_dict = {'asc': 1, 'desc': -1}


class DataTablesServer(object):

    def __init__(self, request, columns, index, db, collection):

        self.columns = columns
        self.index = index
        self.collection = collection
        self.db = db

        # values specified by the datatable for filtering, sorting, paging
        self.request_values = request

        # connection to your mongodb (see pymongo docs).
        # this is defaulted to localhost
        self.dbh = Client("mongodb://{0}:{1}/{2}".format(
                conf['socket_host'],
                conf['socket_port'],
                conf['db']))

        # results from the db
        self.result_data = None

        # total in the table after filtering
        self.cardinality_filtered = 0

        # total in the table unfiltered
        self.cardinality = 0

        self.run_queries()

    def output_result(self):

        try:
            output = {}
            output['sEcho'] = str(int(self.request_values['sEcho']))
            output['iTotalRecords'] = str(self.cardinality)
            output['iTotalDisplayRecords'] = str(self.cardinality_filtered)
            aaData_rows = []

            for row in self.result_data:
                aaData_row = {}
                for i in range(len(self.columns)):
                    value_ = return_value_from_dict(row, self.columns[i])
                    aaData_row[self.columns[i]] = value_
                aaData_rows.append(aaData_row)

            output['aaData'] = aaData_rows

            return output
        except:
            self.dbh.PUT('logs', 'datatable', {'output': _traceback()})
            return []

    def run_queries(self):

        # 'mydb' is the actual name of your database
        mydb = self.dbh[self.db]

        # pages has 'start' and 'length' attributes
        pages = self.paging()

        # the term you entered into the datatable search
        _filter = self.filtering()

        # the document field you chose to sort
        sorting = self.sorting()

        # get result from db to display on the current page
        self.result_data = list(mydb[self.collection].find(filter=_filter,
                                                           skip=pages.start,
                                                           limit=pages.length,
                                                           sort=sorting))

        # length of filtered set
        length_ = len(list(mydb[self.collection].find(filter=_filter)))
        self.cardinality_filtered = length_

        # length of all results you wish to display in the datatable, unfiltered
        self.cardinality = len(list(mydb[self.collection].find()))

    def filtering(self):

        # build your filter spec
        _filter = {}
        check1 = ('sSearch' in self.request_values)
        check2 = (self.request_values['sSearch'] != "")
        if check1 and check2:

            # the term put into search is logically concatenated
            # with 'or' between all columns
            or_filter_on_all_columns = []

            for i in range(len(self.columns)):
                column_filter = {}
                # case insensitive partial string matching pulled from user
                # input
                column_filter[self.columns[i]] = {
                    '$regex': self.request_values['sSearch'],
                    '$options': 'i'
                    }
                or_filter_on_all_columns.append(column_filter)

            _filter['$or'] = or_filter_on_all_columns

        # individual column filtering - uncomment if needed

        #and_filter_individual_columns = []
        #for i in range(len(columns)):
        #    if (request_values.has_key('sSearch_%d' % i) and
        # request_values['sSearch_%d' % i] != ''):
        #        individual_column_filter = {}
        #        individual_column_filter[columns[i]] = {'$regex':
        # request_values['sSearch_%d' % i], '$options': 'i'}
        #        and_filter_individual_columns.append(individual_column_filter)

        #if and_filter_individual_columns:
        #    _filter['$and'] = and_filter_individual_columns

        return _filter

    def sorting(self):

        order = []
        # mongo translation for sorting order
        if ( self.request_values['iSortCol_0'] != "" ) and ( self.request_values['iSortingCols'] > 0 ):

            for i in range( int(self.request_values['iSortingCols']) ):
                # column number
                column_number = int(self.request_values['iSortCol_'+str(i)])
                # sort direction
                sort_direction = self.request_values['sSortDir_'+str(i)]

                order.append((self.columns[column_number], order_dict[sort_direction]))

        return order

    def paging(self):

        pages = namedtuple('pages', ['start', 'length'])

        if (self.request_values['iDisplayStart'] != "" ) and (self.request_values['iDisplayLength'] != -1 ):
            pages.start = int(self.request_values['iDisplayStart'])
            pages.length = int(self.request_values['iDisplayLength'])

        return pages



#  utils.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
from bson import json_util, SON
from traceback import format_exc as _traceback
import datetime
import string
import re
import os
import io

import uuid
from enum import Enum
from typing import List, Optional, Sequence, Type, TypeVar, Union

from bson.objectid import ObjectId
from .vars import TYPES, SCHEMA_TYPES
import inspect
import yaml
import mimetypes
import urllib

import logging

logger = logging.getLogger("cervmongo")

PUNCTUATION_TRANSLATOR = str.maketrans('', '', string.punctuation)
GENERIC_MIMETYPE = "application/octet-stream"

def detect_mimetype(filename) -> str:
    mimetype = 'application/octet-stream'
    try:
        import magic
        mime = magic.Magic(magic_file="bin/magic", mime=True)

        if isinstance(filename, str):
            mimetype = mime.from_file(filename)
        else:
            mimetype = mime.id_buffer(filename)
            filename.seek(0)
    except:
        mimetype = mimetypes.guess_type(filename)[0]

    return mimetype

def file_and_fileobj(fileobj):
    if isinstance(fileobj, str):
        if os.path.exists(fileobj):
            return open(fileobj, 'rb')
        else:
            return fileobj
    elif isinstance(fileobj, (io.TextIOBase, io.BufferedIOBase, io.RawIOBase, io.IOBase)):
        fileobj.seek(0)
        return fileobj
    else:
        raise TypeError("fileobj is not a valid str or file-like obj; received '{}'".format(type(fileobj)))

def get_file_meta_information(fileobj, filename:str=None, content_type:str=None, extension:str=None, **kwargs) -> dict:
    filename = filename or getattr(fileobj, "filename", str(ObjectId()))
    content_type = content_type or getattr(fileobj, "content_type", None)
    extension = extension or getattr(fileobj, "extension", None)

    if not extension:
        if "." in filename:
            extension = ".{}".format(filename.split(".")[-1].lower())
        elif content_type:
            extension = mimetypes.guess_extension(content_type)
        else:
            if os_path.exists(filename):
                content_type = detect_mimetype(filename)
                extension = mimetypes.guess_extension(content_type)
            else:
                content_type = detect_mimetype(fileobj)
                extension = mimetypes.guess_extension(content_type)

    if not content_type:
        if extension:
            content_type = mimetypes.types_map.get(extension, GENERIC_MIMETYPE)
        else:
            content_type = getattr(fileobj, "content_type", detect_mimetype(fileobj))

    if filename and extension:
        filename = filename.lower()
        extension = extension.lower()
        if not extension in filename:
            filename = "{}{}".format(filename, extension)

    assert all((extension, content_type)), "extension and/or content_type not found"

    return {
        "filename": filename,
        "content_type": content_type,
        "extension": extension,
        "aliases": [filename.split(".")[0], filename.upper()]
        }

def dict_to_query(dictionary:dict) -> str:
    return urllib.parse.urlencode(dictionary)

def sort_list(item, field:str):
    try:
        fields = field.split(".")
        total = len(fields) - 1
        for index, field in enumerate(fields):
            if index != total:
                if field.isdigit():
                    item = item[int(field)]
                elif field in item:
                    item = item[field]
                else:
                    return None
            else:
                if field.isdigit():
                    return item[int(field)]
                elif field in item:
                    return item[field]
                else:
                    return None
    except:
        return None

def getenv_boolean(var_name, default_value=False):
    result = default_value
    env_value = os.getenv(var_name)
    if env_value is not None:
        result = env_value.upper() in ("TRUE", "1")
    return result

def current_datetime(alt:str=False) -> datetime.datetime:
    """Returns current datetime object by default. Accepts alternate format for string format result."""
    if alt:
        return datetime.datetime.now().strftime(alt)
    return datetime.datetime.now()

def current_date(alt:str=False) -> datetime.date:
    """Returns current date object by default. Accepts alternate format for string format result."""
    if alt:
        return datetime.date.today().strftime(alt)
    return datetime.date.today()

def clean_kwargs(*, ONLY:list=[], kwargs:dict={}) -> dict:
    """Allows for sanitization of keyword args before passing to another function"""
    if ONLY:
        return {only_key: kwargs.get(only_key, None) for only_key in ONLY if only_key in kwargs}
    else:
        return kwargs

def parse_string_header(string:str) -> str:
    """For use when parsing nested data from tabular data formats, such as spreadsheets"""
    if string.startswith("{"):
        return ".".join(re.findall(r'(\w+)\b', string, re.DOTALL))
    else:
        return string

def format_string_for_id(string:str) -> str:
    """Cleans string to allow for functional, readable, and permissible MongoDB ID"""
    string = string.translate(PUNCTUATION_TRANSLATOR)
    string = string.replace(" ", "")
    return string.lower()

def return_value_from_dict(dictionary:dict, key:str, if_not:str=" "):
    value = dictionary.get(key if key != "__id" else "_id", if_not)

    if isinstance(value, datetime.date):
        value = value.strftime('%Y/%m/%d')
    elif isinstance(value, ObjectId):
        pass
    elif isinstance(value, (set, tuple, list, iter)):
        try:
            value = ", ".join(value)
        except:
            # INFO: silent fail, intentional
            value = value
    else:
        try:
            value = value
        except:
            # INFO: silent fail, intentional
            pass
    if instance(value, str):
        return value.replace('"', '\\"')
    else:
        return value

def snake2camel(snake:str, start_lower:bool=False) -> str:
    """
    Converts a snake_case string to camelCase.
    The `start_lower` argument determines whether the first letter in the generated camelcase should
    be lowercase (if `start_lower` is True), or capitalized (if `start_lower` is False).
    """
    camel = snake.title()
    camel = re.sub("([0-9A-Za-z])_(?=[0-9A-Z])", lambda m: m.group(1), camel)
    if start_lower:
        camel = re.sub("(^_*[A-Z])", lambda m: m.group(1).lower(), camel)
    return camel


def camel2snake(camel:str) -> str:
    """
    Converts a camelCase string to snake_case.
    """
    snake = re.sub(r"([a-zA-Z])([0-9])", lambda m: f"{m.group(1)}_{m.group(2)}", camel)
    snake = re.sub(r"([a-z0-9])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    return snake.lower()


def objectid_representer(dumper, data):
    return dumper.represent_scalar("!_id", str(data))

def objectid_constructor(loader, data):
    return ObjectId(loader.construct_scalar(data))

yaml.SafeDumper.add_representer(ObjectId, objectid_representer)
yaml.add_constructor('!_id', objectid_constructor)

def _get_class_that_defined_method(meth):
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        if isinstance(cls, type):
            return cls
    return None

def generate_new_id() -> str:
    return str(uuid.uuid4())

def ensure_enums_to_strs(items: Union[Sequence[Union[Enum, str]], Type[Enum]]):
    str_items = []
    for item in items:
        if isinstance(item, Enum):
            str_items.append(str(item.value))
        else:
            str_items.append(str(item))
    return str

def yaml_dump(data:dict) -> str:
    return yaml.safe_dump(data, default_flow_style=False)

def yaml_load(data, _file:bool=False) -> dict:
    if _file:
        return yaml.load(open(data, 'r'))
    else:
        return yaml.safe_load(data)

def json_dump(data:dict, pretty:bool=False) -> str:
    if pretty:
        return json_util.dumps(data, indent=4, sort_keys=True)
    else:
        return json_util.dumps(data)

def json_load(data:str) -> dict:
    return json_util.loads(data)

def clean_traceback() -> str:
    traceback = _traceback
    # TODO: cleaning logic, to dict, maybe make class?
    return traceback

def silent_drop_kwarg(kwargs:dict, key:str, reason:str=""):
    if reason:
        logger.debug(f"dropping key {key} for reason {reason}")
    else:
        logger.debug(f"dropping key {key}")
    return kwargs.pop(key)

# TODO: custom jsonschema validator


# INFO: tools to use with JSON samples
def type_from_schema(schema_type:str):
    """retrieves type function based on JSON sample inferred data schema type"""
    schema_type = schema_type.lower()
    return TYPES[schema_type]


def schema_from_dict(dictionary:dict, additional:dict={}):
    """creates a simple JSON schema from JSON sample document"""
    schema = {"type": "object", "required": [], "properties" : {}}

    for key, value in dictionary.items():
        if ":" in key:
            key, _type = key.strip().split(":")
            _type = SCHEMA_TYPES[_type]
        else:
            if isinstance(value, str):
                _type = "str"
            elif isinstance(value, float):
                _type = "float"
            elif isinstance(value, int):
                _type = "int"
            elif isinstance(value, dict):
                _type = "dict"
            elif isinstance(value, ObjectId):
                _type = "oid"
            elif isinstance(value, datetime.datetime):
                _type = "datetime"
            elif isinstance(value, datetime.date):
                _type = "date"
            elif isinstance(value, bool):
                _type = "bool"
            else:
                raise TypeError("unrecognized type '{}' for value '{}'".format(type(value), value))
            _type = SCHEMA_TYPES[_type]

        required = False
        if key.endswith("*"):
            key = key.strip("*")
            required = True

        schema["properties"][key] = {
                "type": _type
            }

        if required:
            if not key in schema["required"]:
                schema["required"].append(key)

    return schema



#  cervmongo
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
__all__ = [
        "get_client", "get_doc", "SUPPORT_GRIDFS",
        "get_async_client", "get_async_doc", "SUPPORT_ASYNCIO_CLIENT", "SUPPORT_ASYNCIO_BUCKET",
        "SUPPORT_PYDANTIC",
        "ASC", "DESC",
        "quick_load_client",
        "get_config",
        ]

from . import (
            main,
            aio,
            models,
            vars,
            config,
            extra,
            )
from .extra import convenience as convenience

quick_load_client = convenience.quick_load_client

get_client = main.get_client
get_doc = main.get_doc
SUPPORT_GRIDFS = main.SUPPORT_GRIDFS

get_async_client = aio.get_async_client
get_async_doc = aio.get_async_doc
SUPPORT_ASYNCIO_CLIENT = aio.SUPPORT_ASYNCIO_CLIENT
SUPPORT_ASYNCIO_BUCKET = aio.SUPPORT_ASYNCIO_BUCKET

SUPPORT_PYDANTIC = models.SUPPORT_PYDANTIC

ASC = vars.ASCENDING
DESC = vars.DESCENDING

config = config.Config

def get_config() -> config:
    """returns the Config class to set MongoDB configuration"""
    return config



#  main.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
__all__ = ["SUPPORT_GRIDFS", "get_client", "get_doc", "SyncIOClient", "SyncIODoc"]

from os import path as os_path
from pymongo import MongoClient
from pymongo import WriteConcern
from dateutil.parser import parse as dateparse
import types
from jsonschema import validate
from functools import partial
import typing
import copy

from .models import (
                GenericResponse,
                StandardResponse,
                MongoListResponse,
                MongoDictResponse,
                )
from .vars import (
                StringEnum,
                IntEnum,
                PAGINATION_SORT_FIELDS,
                ENUM,
                DOC_ID,
                )
from .utils import (
                get_file_meta_information,
                parse_string_header,
                format_string_for_id,
                silent_drop_kwarg,
                current_datetime,
                file_and_fileobj,
                detect_mimetype,
                dict_to_query,
                clean_kwargs,
                current_date,
                json_load,
                json_dump,
                logger,
                )
from .config import Config

try:
    from gridfs import GridFSBucket
    SUPPORT_GRIDFS = True #: True if gridfs functionality is available else False
except:
    logger.warning("gridfs is not installed")
    SUPPORT_GRIDFS = False #: True if gridfs functionality is available else False
    class GridFSBucket: pass # NOTE: in case of refereneces

_CollectionClientType = typing.TypeVar("CollectionClient")


class SyncIOClient(MongoClient):
    """
        High-level MongoClient subclass with additional methods added for ease-of-use,
        having some automated conveniences and defaults.
    """
    _MONGO_URI = lambda _: getattr(Config, "MONGO_URI", None) #: Valid MongoDB URI, defaults to Config.MONGO_URI if not supplied
    _DEFAULT_COLLECTION = None
    _KWARGS = None #: saves originally supplied kwargs, if a reload is required
    _LOGGING_COND_GET = None
    _LOGGING_COND_POST = None
    _LOGGING_COND_PUT = None
    _LOGGING_COND_PATCH = None
    _LOGGING_COND_DELETE = None

    def __init__(self, mongo_uri:typing.Optional[str]=None, default_collection:typing.Optional[str]=None, **kwargs):
        self._MONGO_URI = mongo_uri or self._MONGO_URI
        if callable(self._MONGO_URI):
            self._MONGO_URI = self._MONGO_URI()
        self._DEFAULT_COLLECTION = default_collection or self._DEFAULT_COLLECTION

        if kwargs:
            self._KWARGS = kwargs.copy()

        for kwarg in kwargs.keys():
            if kwarg.lower() in ('logging_cond_get', 'logging_cond_post',
                                'logging_cond_put', 'logging_cond_patch',
                                'logging_cond_delete'):
                setattr(self, kwarg.upper(), kwargs.pop(kwarg))

        MongoClient.__init__(self, self._MONGO_URI, **kwargs)

        db = self.get_default_database()
        if not getattr(db, "name", None):
            logger.warning("database not provided in MONGO_URI, assign with method set_database")
            logger.warning("gridfsbucket not instantiated due to missing database")
        else:
            global SUPPORT_GRIDFS
            if SUPPORT_GRIDFS:
                logger.debug("gridfsbucket instantiated under self.FILES")
                self.FILES = GridFSBucket(db)
            else:
                logger.warning("gridfsbucket not instantiated due to missing 'gridfs' package")
                self.FILES = None

    def __repr__(self):
        db = self.get_default_database()

        if not getattr(db, "name", None) or db.name == "None":
            return "<cervmongo.SyncIOClient>"
        else:
            return f"<cervmongo.SyncIOClient.{db.name}>"

    def _process_record_id_type(self, record):
        one = False
        if isinstance(record, str):
            one = True
            if "$oid" in record:
                record = json_load(record)
            else:
                try:
                    record = DOC_ID.__supertype__(record)
                except:
                    pass
        elif isinstance(record, DOC_ID.__supertype__):
            record = record
            one = True
        elif isinstance(record, dict):
            keys = record.keys()
            if any(["$oid" in keys, "$regex" in keys]):
                record = json_dump(record)
                record = json_load(record)
                one = True
        return (record, one)

    def set_database(database:str) -> None:
        """
            is used to change or set database of client instance, also changes db in Config class
        """
        Config.set_mongo_db(database)
        if self._KWARGS:
            SyncIOClient.__init__(self, mongo_uri=Config.MONGO_URI, default_collection=self._DEFAULT_COLLECTION, **self._KWARGS)
        else:
            SyncIOClient.__init__(self, mongo_uri=Config.MONGO_URI, default_collection=self._DEFAULT_COLLECTION)

    def COLLECTION(self, collection:str):
        """
            returns unique `CollectionClient <#cervmongo.main.cervmongo.main.SyncIOClient.COLLECTION.CollectionClient>`_ class instance

            Be aware, CollectionClient is NOT a valid MongoClient. To access the original
            SyncIOClient instance, use method get_client of CollectionClient instance.
        """

        self._DEFAULT_COLLECTION = collection


        class CollectionClient:
            """Convenience class that auto-supplies collection to all upper-cased SyncIOClient methods, where required"""

            __parent__ = CLIENT = self #! the original SyncIOClient instance
            # INFO: variables
            _DEFAULT_COLLECTION = collection #: the default collection assigned
            _MONGO_URI = self._MONGO_URI #: the MongoDB URI supplied from SyncIOClient instance
            # INFO: general methods
            GENERATE_ID = self.GENERATE_ID
            COLLECTION = self.COLLECTION
            # INFO: GridFS file operations
            UPLOAD = self.UPLOAD
            DOWNLOAD = self.DOWNLOAD
            ERASE = self.ERASE
            # INFO: truncated Collection methods
            INDEX = partial(self.INDEX, collection)
            ADD_FIELD = partial(self.ADD_FIELD, collection)
            REMOVE_FIELD = partial(self.REMOVE_FIELD, collection)
            DELETE = partial(self.DELETE, collection)
            GET = partial(self.GET, collection)
            POST = partial(self.POST, collection)
            PUT = partial(self.PUT, collection)
            PATCH = partial(self.PATCH, collection)
            REPLACE = partial(self.REPLACE, collection)
            SEARCH = partial(self.SEARCH, collection)
            PAGINATED_QUERY = partial(self.PAGINATED_QUERY, collection)
            def __repr__(s):
                return "<cervmongo.SyncIOClient.CollectionClient>"
            def get_client(s):
                return s.CLIENT

        # ~ if function: # NOTE: Make CollectionClient class a function attr?
            # ~ setattr(function, "CollectionClient", CollectionClient) # NOTE: Make CollectionClient class a function attr?

        return CollectionClient()
    # ~ COLLECTION.__defaults__ = COLLECTION.__defaults__[:-1] + (COLLECTION,) # NOTE: Make CollectionClient class a function attr?

    def PAGINATED_QUERY(self, collection:typing.Optional[str], limit:int=20,
                                sort:PAGINATION_SORT_FIELDS=PAGINATION_SORT_FIELDS["_id"],
                                after:str=None, before:str=None,
                                page:int=None, endpoint:str="/",
                                ordering:int=-1, query:dict={}, **kwargs):
        """
            Returns paginated results of cursor from the collection query.

            Available pagination methods:
             - **Cursor-based (default)**
                - after
                - before
                - limit (results per page, default 20)
             - **Time-based** (a datetime field must be selected)
                - sort (set to datetime field)
                - after (records after this time)
                - before (records before this time)
                - limit (results per page, default 20)
             - **Offset-based** (not recommended)
                - limit (results per page, default 20)
                - page
        """
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        if isinstance(sort, ENUM.__supertype__):
            sort = sort.value

        total_docs = self.GET(collection, query, count=True, empty=0)

        if not page:
            if sort == "_id":
                pagination_method = "cursor"
            else:
                pagination_method = "time"
            results = self.GET(collection, query,
                                    limit=limit, key=sort, before=before,
                                    after=after, sort=ordering, empty=[]).list()

        else:
            assert page >= 1, "page must be equal to or greater than 1"
            pagination_method = "offset"
            results = self.GET(collection, query,
                                    perpage=limit, key=sort, page=page,
                                    sort=ordering, empty=[]).list()

        # INFO: determine 'cursor' template
        if sort == "_id":
            template = "_{_id}"
        else:
            template = "{date}_{_id}"

        new_after = None
        new_before = None

        if results:
            _id = results[-1]['_id']
            try:
                date = results[-1][sort].isoformat()
            except:
                date = None
            if len(results) == limit:
                new_after = template.format(_id=_id, date=date)

            _id = results[0]['_id']
            try:
                date = results[0][sort].isoformat()
            except:
                date = None
            if any((after, before)):
                new_before = template.format(_id=_id, date=date)

            if pagination_method in ("cursor", "time"):
                if before:
                    check_ahead = self.GET(collection, query,
                                            limit=limit, key=sort, before=new_before, count=True, empty=0)
                    if not check_ahead:
                        new_before = None
                elif after:
                    check_ahead = self.GET(collection, query,
                                            limit=limit, key=sort, after=new_after, count=True, empty=0)
                    if not check_ahead:
                        new_after = None



        response = {
            "data": results,
            "details": {
                "pagination_method": pagination_method,
                "query": dict_to_query(query),
                "sort": sort,
                "unique_id": getattr(self, "_UNIQUE_ID", "_id"),
                "total": total_docs,
                "count": len(results),
                "limit": limit
                }
            }

        endpoint = endpoint

        # TODO: Refactor
        if pagination_method in ("cursor", "time"):
            response["details"]["cursors"] = {
                  "after": new_after,
                  "before": new_before
                }
            before_url_template = "{endpoint}?sort={sort}&limit={limit}&before={before}"
            after_url_template = "{endpoint}?sort={sort}&limit={limit}&after={after}"
        else: # INFO: pagination_method == "offset"
            response["details"]["cursors"] = {
                  "prev_page": page - 1 if page > 1 else None,
                  "next_page": page + 1 if (page * limit) <= total_docs else None
                }
            before_url_template = "{endpoint}?sort={sort}&limit={limit}&page={page}"
            after_url_template = "{endpoint}?sort={sort}&limit={limit}&page={page}"

        if new_before:
            response["details"]["previous"] = before_url_template.format(
                                                                    endpoint=endpoint,
                                                                    sort=sort,
                                                                    page=page,
                                                                    limit=limit,
                                                                    after=new_after,
                                                                    before=new_before)
        else:
            response["details"]["previous"] = None

        if new_after:
            response["details"]["next"] = after_url_template.format(
                                                                    endpoint=endpoint,
                                                                    sort=sort,
                                                                    page=page,
                                                                    limit=limit,
                                                                    after=new_after,
                                                                    before=new_before)
        else:
            response["details"]["next"] = None

        return response
    PAGINATED_QUERY.clean_kwargs = lambda kwargs: clean_kwargs(ONLY=("limit", "sort", "after",
                                            "before", "page", "endpoint", "query"), kwargs=kwargs)

    def GENERATE_ID(self, _id:str=None) -> DOC_ID:
        """
            returns a unique ObjectID, simply for convenience
        """
        if _id:
            return DOC_ID.__supertype__(_id)
        else:
            return DOC_ID.__supertype__()

    def DELETE(self, collection:typing.Optional[str], record_or_records, soft:bool=False) -> typing.Union[MongoDictResponse, MongoListResponse]:
        """
            deletes the requested document(s)

            returns MongoDB response document

            If soft=true, creates collection ('deleted.{collection}')
            and inserts the deleted document there. field 'oid' is
            guaranteed to equal the original document's '_id'.
        """
        db = self.get_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"

        collection = db[collection]

        record_or_records = self._process_record_id_type(record_or_records)[0]

        if isinstance(record_or_records, (str, DOC_ID.__supertype__)):
            record_or_records = {"_id": record_or_records}

        if isinstance(record_or_records, dict):
            data_record = collection.find_one_and_delete(record_or_records)
            if soft:
                try:
                    data_record["oid"] = data_record["_id"]
                    self.PUT("deleted."+collection, data_record)
                except:
                    data_record["oid"] = data_record.pop("_id")
                    self.PUT("deleted."+collection, data_record)
            return MongoDictResponse(data_record)
        elif isinstance(record_or_records, (list, tuple)):
            results = []
            for _id in record_or_records:
                data_record = collection.find_one_and_delete({'_id': _id})
                if soft:
                    try:
                        data_record["oid"] = data_record["_id"]
                        self.PUT("deleted."+collection, data_record)
                    except:
                        data_record["oid"] = data_record.pop("_id")
                        self.PUT("deleted."+collection, data_record)
                results.append(data_record)

            return MongoListResponse(results)
        else:
            raise TypeError("record_or_records was of invalid type '{}'".format(type(record_or_records)))

    def INDEX(self, collection, key:str="_id", sort:int=1, unique:bool=False, reindex:bool=False) -> None:
        """
            creates an index, however most useful in constraining certain fields as unique
        """
        db = self.get_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        if reindex:
            collection.reindex()
        else:
            name = "%sIndex%s" % (key, "Asc" if sort == 1 else "Desc")
            try:
                if not name in collection.index_information():
                    collection.create_index([
                        (key, sort)], name=name, background=True, unique=unique)
                try:
                    # INFO: a full text-index is recommended when no fixed schema / document structure is in place
                    collection.create_index([("$**", "text")], name="textIndex", background=True)
                except:
                    pass
            except:
                # TODO: provide a short clear exception?
                raise

    def ADD_FIELD(self, collection, field:str, value:typing.Union[typing.Dict, typing.List, str, int, float, bool]='', data=False, query:dict={}) -> None:
        """
            adds field with value provided to all records in collection that match query

            kwarg 'data', if provided, is an existing field in the record that can be used as the default value.
            useful for name changes of fields in schema.
        """
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        query.update({field: {'$exists': False}})
        if data:
            records = self.GET(collection, query, fields={
                data: True}, empty=[])
        else:
            records = self.GET(collection, query, fields={
                '_id': True}, empty=[])

        for record in records:
            if data:
                self.PATCH(collection, record['_id'], {"$set": {
                    field: record[data]}})
            else:
                self.PATCH(collection, record['_id'], {"$set": {
                    field: value}})

    def REMOVE_FIELD(self, collection, field:str, query:dict={}) -> None:
        """
            removes field of all records in collection that match query

            useful for complete field removal, once field is no longer needed.
        """
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        query.update({field: {'$exists': True}})
        records = self.GET(collection, query, distinct=True)

        for record in records:
            self.PATCH(collection, record, {"$unset": {field: ""}})

    def UPLOAD(self, fileobj:typing.Union[typing.IO, str], filename:str=None, content_type:str=None, extension:str=None, **kwargs):
        """
            returns GridFS response document after successful upload

            fileobj can be either valid filepath or file/file-like object,
            all other kwargs are stored as meta data
        """
        assert self.FILES, "GridFS instance not initialized, run method 'set_database' with the desired database and try again"
        fileobj = file_and_fileobj(fileobj)
        metadata = get_file_meta_information(fileobj, filename=filename, content_type=content_type, extension=extension)
        filename = metadata['filename']
        metadata.update(kwargs)
        return self.FILES.upload_from_stream(filename, fileobj, metadata=metadata)

    def ERASE(self, filename_or_id:typing.Union[str, DOC_ID], revision:int=-1) -> None:
        """
            deletes the GridFS file. if multiple revisions, deletes most recent by default.

            filename_or_id can be the _id field value or the filename value.
        """
        assert self.FILES, "GridFS instance not initialized, run method 'set_database' with the desired database and try again"
        fs_doc = self.DOWNLOAD(filename_or_id, revision=revision)
        self.FILES.delete(fs_doc._id)
        fs_doc.close()

    def DOWNLOAD(self, filename_or_id:typing.Optional[typing.Union[str, DOC_ID]]=None, revision:int=-1, skip:int=None, limit:int=None, sort:int=-1, **query):
        """
            returns download stream of file if filename_or_id is provided, else returns a cursor of files matching query
        """
        assert self.FILES, "GridFS instance not initialized, run method 'set_database' with the desired database and try again"
        revision = int(revision)
        if filename_or_id:
            if isinstance(filename_or_id, DOC_ID.__supertype__):
                return self.FILES.open_download_stream(filename_or_id)
            else:
                return self.FILES.open_download_stream_by_name(filename_or_id, revision=revision)

        return self.FILES.find(query, limit=limit, skip=skip, sort=sort, no_cursor_timeout=True)

    def GET(self, collection, id_or_query:typing.Union[DOC_ID, typing.Dict, str]={}, sort:int=1, key:str="_id", count:bool=None, search:str=None, fields:dict=None, page:int=None, perpage:int=False, limit:int=None, after:str=None, before:str=None, empty=None, distinct:str=None, one:bool=False, **kwargs):
        """
            record can be either _id (accepts unicode form of ObjectId, as well as extended JSON bson format) or query

            Unless certain kwargs are used (count, distinct, one), behaviour is as follows:

            - if _id is recognized, one is set to true and will return the exact matching document
            - if query is recognized, will return MongoList response of cursor
            - if count is provided and _id is not recognized, returns number of documents in cursor
            - if distinct is provided, returns a unique list of the field values (accepts dot notation)
            - if one is provided, returns the first matching document of cursor

            kwargs count, distinct, one cannot be used together, priority is as follows if all are provided:

            1. count
            2. distinct
            3. one
        """
        db = self.get_default_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection not provided"

        if not isinstance(collection, (list, tuple, types.GeneratorType)):
            collection = [collection]
        cols = list(set(collection))
        results = []
        number_of_results = len(cols)

        if distinct == True:
            distinct = "_id"

        id_or_query, _one = self._process_record_id_type(id_or_query)
        one = _one if _one else one
        if _one:
            query = {"_id": id_or_query}
        else:
            query = id_or_query

        for collection in cols:
            collection = db[collection]

            if query or not search:
                if count and not limit:
                    if query:
                        results.append(collection.count_documents(query, **kwargs))
                    else:
                        results.append(collection.estimated_document_count(**kwargs))
                elif distinct:
                    cursor = collection.find(query, **kwargs)
                    results.append(cursor.sort([(key, sort)]).distinct(distinct))
                elif perpage:
                    total = (page - 1) * perpage
                    cursor = collection.find(query, fields, **kwargs)
                    results.append(MongoListResponse(cursor.sort([(key, sort)]).skip(total).limit(perpage)))
                elif limit:
                    if any((query, after, before)):
                        query = {"$and": [
                                    query
                                ]}
                    if after or before:
                        if after:
                            sort_value, _id_value = after.split("_")
                            _id_value = DOC_ID.__supertype__(_id_value)
                            query["$and"].append({"$or": [
                                            {key: {"$lt": _id_value}}
                                        ]})
                            if key != "_id":
                                sort_value = dateparse(sort_value)
                                query["$and"][-1]["$or"].append({key: {"$lt": sort_value}, "_id": {"$lt": _id_value}})
                        elif before:
                            sort_value, _id_value = before.split("_")
                            _id_value = DOC_ID.__supertype__(_id_value)
                            query["$and"].append({"$or": [
                                            {key: {"$gt": _id_value}}
                                        ]})
                            if key != "_id":
                                sort_value = dateparse(sort_value)
                                query["$and"][-1]["$or"].append({key: {"$gt": sort_value}, "_id": {"$gt": _id_value}})

                    if count:
                        try:
                            cursor = collection.find(query, fields, **kwargs).sort([(key, sort)]).limit(limit).count(with_limit_and_skip=True)
                        except:
                            cursor = collection.count_documents(query, limit=limit, hint=[(key, sort)], **kwargs)
                        results.append(cursor)
                    else:
                        cursor = collection.find(query, fields, **kwargs).sort([(key, sort)]).limit(limit)
                        results.append(MongoListResponse(cursor))
                elif one:
                    val = collection.find_one(query, fields, **kwargs)
                    results.append(MongoDictResponse(val) if val else empty)
                else:
                    cursor = collection.find(query, fields, **kwargs).sort([(key, sort)])
                    results.append(MongoListResponse(cursor))
            elif search:
                try:
                    cursor = collection.find({"$text": {"$search": search}})
                    if count:
                        results.append(cursor.count())
                    elif distinct:
                        results.append(cursor.distinct(distinct))
                    if perpage:
                        total = (page - 1) * perpage
                        results.append(MongoListResponse(cursor.sort([(key, sort)]).skip(total).limit(perpage)))
                    results.append(MongoListResponse(cursor.sort([(key, sort)])))
                except:
                    cursor = collection.command('textIndex', search=search)
                    if count:
                        results.append(cursor.count())
                    elif distinct:
                        results.append(cursor.distinct(distinct))
                    if perpage:
                        total = (page - 1) * perpage
                        results.append(MongoListResponse(cursor.sort([(key, sort)]).skip(total).limit(perpage)))
                    results.append(MongoListResponse(cursor.sort([(key, sort)])))
            else:
                raise Error("unidentified error")

        if number_of_results == 1:
            return results[0]
        else:
            return results

    def SEARCH(self, collection, search:str, **kwargs):
        """
            returns the results of querying textIndex of collection
        """
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        return self.GET(collection, search=search, **kwargs)

    def POST(self, collection, record_or_records:typing.Union[typing.List, typing.Dict]):
        """
            creates new record(s) and returns MongoDB response document
        """
        db = self.get_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        if isinstance(record_or_records, (list, tuple)):
            return collection.insert_many(record_or_records)
        elif isinstance(record_or_records, dict):
            return collection.insert_one(record_or_records)
        else:
            raise TypeError("invalid record_or_records type '{}' provided".format(type(record_or_records)))

    def PUT(self, collection, record_or_records:typing.Union[typing.List, typing.Dict]):
        """
            creates or replaces record(s) with exact _id provided, _id is required with record object(s)

            returns original document, if replaced
        """
        db = self.get_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        results = []

        if isinstance(record_or_records, (list, tuple)):
            assert all([ record.get("_id", None) for record in record_or_records ]), "not all records provided contained an _id"
            return collection.insert_many(record_or_records)
        elif isinstance(record_or_records, dict):
            assert record_or_records.get("_id", None), "no _id provided"
            query = {"_id": record_or_records["_id"]}
            return collection.find_one_and_replace(query, record_or_records)
        else:
            raise TypeError("invalid record_or_records type '{}' provided".format(type(record_or_records)))

    def REPLACE(self, collection, original, replacement, upsert:bool=False):
        db = self.get_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        return collection.replace_one({'_id': original},
                    replacement, upsert=upsert)

    def PATCH(self, collection, original, updates, w:int=1, upsert:bool=False, multi:bool=False, log=None):
        db = self.get_database()
        collection = collection or self._DEFAULT_COLLECTION
        assert collection, "collection must be of type str"
        collection = db[collection]

        if w != 1:
            WRITE = WriteConcern(w=0)
            _collection = collection.with_options(write_concern=WRITE)
        else:
            _collection = collection
        if log:
            try:
                pass # TODO: optional, may not incorporate
            except:
                pass

        if multi:
            results = _collection.update_many(original, updates, upsert=False) # explictly false
            return results
        elif isinstance(original, (str, DOC_ID.__supertype__)):
            original = self._process_record_id_type(original)
            query = {'_id': original}
            set_on_insert_id = {"$setOnInsert": query}
            updates.update(set_on_insert_id)

            results = _collection.update_one(query, updates, upsert=upsert)
            return results
        else:
            results = []
            for i, _id in enumerate(original):
                _id = self._process_record_id_type(_id)
                query = {'_id': _id}
                set_on_insert_id = {"$setOnInsert": query}
                updates[i].update(set_on_insert_id)

                result = collection.update_one(query, updates[i], upsert=upsert)
                results.append(result)
            db["_logs"].insert_one({
                                    'name': 'reindex',
                                    'db': db,
                                    'collection': collection,
                                    'datetime': current_datetime()
                                    })
            return results


class SyncIODoc(SyncIOClient):
    """
        Custom MongoClient subclass with customizations for creating
        standardized documents and adding json schema validation.
    """
    _MONGO_URI = lambda _: getattr(Config, "MONGO_URI", None)
    _DOC_TYPE:str = None #: MongoDB collection to use
    _DOC_ID:str = "_id" #:
    _DOC_SAMPLE:str = None
    _DOC_SCHEMA:str = None
    _DOC_MARSHMALLOW:str = False
    _DOC_DEFAULTS:dict = {}
    _DOC_RESTRICTED_KEYS:list = []
    _DOC_ENUMS:list = []
    _DOC_SETTINGS:str = "settings"

    def __init__(self, _id=None, doc_type:str=None, doc_sample:typing.Union[typing.Dict, str]=None, doc_schema:typing.Union[typing.Dict, str]=None, doc_id:str=None, mongo_uri:str=None, **kwargs):
        self._MONGO_URI = mongo_uri or self._MONGO_URI
        if callable(self._MONGO_URI):
            self._MONGO_URI = self._MONGO_URI()
        # INFO: set default collection
        self._DOC_TYPE = doc_type or self._DOC_TYPE
        assert self._DOC_TYPE, "collection must be of type str"
        self._DEFAULT_COLLECTION = self._DOC_TYPE
        # INFO: location for sample record, used as template
        self._DOC_SAMPLE = doc_sample or self._DOC_SAMPLE
        # INFO: path to validation schema
        self._DOC_SCHEMA = doc_schema or self._DOC_SCHEMA
        # INFO: sets the unique id field for the document, if any (cannot be _id)
        self._DOC_ID = doc_id or self._DOC_ID
        assert self._DOC_ID, "unique id field name must be of type str"

        for kwarg in kwargs.keys():
            if kwarg.lower() in ('doc_marshmallow', 'doc_defaults', 'doc_restricted_keys', "doc_enums"):
                setattr(self, "_{}".format(kwarg.upper()), kwargs.pop(kwarg))

        # Initial Record object with sample else start blank dict
        if self._DOC_SAMPLE:
            if isinstance(self._DOC_SAMPLE, str):
                sample_full_path = os_path.join(Config.JSON_SAMPLE_PATH, self._DOC_SAMPLE)
                with open(sample_full_path) as file1:
                    sample = json_load(file1.read())
            elif isinstance(self._DOC_SAMPLE, dict):
                sample = self._DOC_SAMPLE
            else:
                raise TypeError("_DOC_SAMPLE is invalid type '{}', valid types are dict and str".format(type(self._DOC_SAMPLE)))
            sample_parent_found = sample.pop("__parent__", None)
            while sample_parent_found:
                parent_sample_full_path = os_path.join(Config.JSON_SAMPLE_PATH, sample_parent_found)
                with open(parent_sample_full_path) as _file:
                    parent_sample = json_load(_file.read())
                parent_sample.update(sample)
                sample = parent_sample
                sample_parent_found = sample.pop("__parent__", None)
            self.sample = sample
        else:
            self.sample = {}

        # INFO: Load schema else start blank dict to add manual validation entries
        if self._DOC_SCHEMA:
            if isinstance(self._DOC_SCHEMA, str):
                schema_full_path = os_path.join(Config.JSON_SCHEMA_PATH, self._DOC_SCHEMA)
                with open(schema_full_path) as _file:
                    self.schema = json_load(_file.read())
            elif isinstance(self._DOC_SCHEMA, dict):
                self.schema = self._DOC_SCHEMA
            else:
                raise TypeError("_DOC_SCHEMA is invalid type '{}', valid types are dict and str".format(type(self._DOC_SCHEMA)))
        else:
            self.schema = {}

        SyncIOClient.__init__(self, **kwargs)

        db = self.get_default_database()

        if not getattr(db, "name", None) or db.name == "None":
            raise Exception("database not provided in MongoDB URI")
        else:
            self._DOC_DB = db.name

        # Initialize enums
        if not self._DOC_ENUMS:
            pass
            #enums_record = self.GET(self._DOC_SETTINGS, "enums"

        # INFO: If class has a _DOC_ID assigned, create unique index
        if self._DOC_ID != "_id":
            self.INDEX(self._DOC_TYPE, key=self._DOC_ID,
                                                        sort=1, unique=True)

        self.load(_id)

    def __repr__(self):
        if self.RECORD.get("_id", None):
            _id = self.id()
            return f"<cervmongo.SyncIODoc.{self._DOC_DB}.{self._DOC_TYPE}.{self._DOC_ID}:{_id}>"
        else:
            return "<cervmongo.SyncIODoc>"

    def __enter__(self):
        return self

    def __exit__(self):
        self.close()

    def _process_restrictions(self, record:dict=None):
        """removes restricted keys from record and return record"""
        try:
            if record:
                assert isinstance(record, (MongoDictResponse, dict)), "Needs to be a dictionary, got {}".format(type(record))
                return {key: value for key, value in record.items() if not key in self._DOC_RESTRICTED_KEYS}
            else:
                return {key: value for key, value in self.RECORD.items() if not key in self._DOC_RESTRICTED_KEYS}
        except:
            logger.exception("encountered error when cleaning self.RECORD, returning empty dict")
            return {}

    def _p_r(self, record:dict=None):
        """truncated alias for _process_restrictions"""
        return self._process_restrictions(record=record)

    def _generate_unique_id(self, template:str="{total}", **kwargs):
        return template.format(**kwargs).upper()

    def _timestamp(self, value:str=None):
        if value:
            try:
                value = dateparse(value)
            except:
                value = current_datetime()
        else:
            value = current_datetime()
        return value

    def _guess_corresponding_fieldname(self, _type:str="unknown", related_field:str=""):
        time_fields = ("date", "datetime", "time")
        if _type in time_fields:
            # NOTE: a timestamp is 'mostly' accompanied by a user or relation
            if related_field:
                if "_" in related_field:
                    field_parts = related_field.split("_")
                elif "-" in related_field:
                    field_parts = related_field.split("-")
                else:
                    field_parts = related_field.split()

                for field_part in field_parts:
                    field_part = field_part.strip(" _-").lower()
                    if any(x in field_part for x in time_fields):
                        continue
                    else:
                        return f"{field_part}_by"
                return "for"
            else:
                return "by"
        else:
            # NOTE: an unknown type field has a timestamp pairing or desc
            if related_field:
                return f"{related_field}_description"
            else:
                return "field_description"

    def _related_record(self, collection=None, field:str="_id", value=False, additional:dict={}):
        additional.update({
            field: True
            })

        record = self.GET(collection, {field: value}, fields=additional, one=True, empty={})
        assert record, 'Error: No record found'

        record['key'] = field
        record['collection'] = collection
        record[field] = record[field]

        return record

    def load(self, _id=None):
        # If _id specified on init, load actual record versus blank template
        if _id:
            if self._DOC_ID:
                self.RECORD = self.GET(self._DOC_TYPE, {self._DOC_ID: _id}, one=True)
            else:
                self.RECORD = self.GET(self._DOC_TYPE, _id)
        else:
            self.RECORD = copy.deepcopy(self.sample)

        if not self.RECORD:
            self.RECORD = {}

        return StandardResponse(
                    data=self._p_r(self.RECORD),
                    details={
                "state": "unsaved" if not self.RECORD.get("_id", None) else "saved",
                "unique_id": self._DOC_ID
                })

    def view(self, _id=False):
        if not _id:
            return StandardResponse(data=self._p_r(self.RECORD))
        else:
            if self._DOC_ID:
                return StandardResponse(
                            data=self._p_r(self.GET(self._DOC_TYPE, {self._DOC_ID: _id}, one=True, empty={})),
                            details={
                        "unique_id": self._DOC_ID
                        }
                    )
            else:
                return StandardResponse(
                            data=self._p_r(self.GET(self._DOC_TYPE, {"_id": _id}, one=True, empty={})),
                            details={
                        "unique_id": self._DOC_ID
                        }
                    )

    def reload(self):
        assert self.RECORD.get('_id')
        self.RECORD = self.GET(self._DOC_TYPE, self.RECORD['_id'])

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "unique_id": self._DOC_ID
                }
            }

    def id(self):
        return self.RECORD.get(self._DOC_ID, None)

    def create(self, save:bool=False, trigger=None, template:str="{total}", query:dict={}, **kwargs):
        assert self.RECORD.get('_id') is None, """Cannot use create method on
 an existing record. Use patch method instead."""

        if self._DOC_MARSHMALLOW:
            self._DOC_MARSHMALLOW().load(kwargs)
            self.RECORD.update(kwargs)
        elif self.sample:
            # INFO: removing invalid keys based on sample record
            [ silent_drop_kwarg(kwargs, x, reason="not in self.sample") for x in list(kwargs.keys()) if not x in self.sample ]
            self.RECORD.update(kwargs)
        else:
            self.RECORD.update(kwargs)

        kwargs['total'] = str(self.GET(self._DOC_TYPE, query).count() + 1).zfill(6)

        if self._DOC_ID and not self.RECORD.get(self._DOC_ID):
            self.RECORD[self._DOC_ID] = self._generate_unique_id(template=template, **kwargs)

        if save:
            self.save(trigger=None)

        return {
            "data": self._p_r(self.RECORD),
            "details": {"unique_id": self._DOC_ID, "collection": self._DOC_TYPE, "_id": self.RECORD[self._DOC_ID]}
            }

    def push(self, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use push method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        self.PATCH(None, self.RECORD['_id'], {"$push": kwargs})
        self.reload()

        keys = list(kwargs.keys())
        values = [ kwargs[key] for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "push",
                "desc": "push a value to end of array, not set",
                "field": kwargs.keys(),
                "value": kwargs.values()
                }
            }

    def pull(self, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use pull method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        self.PATCH(None, self.RECORD['_id'], {"$pull": kwargs})
        self.reload()

        keys = list(kwargs.keys())
        values = [ kwargs[key] for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "pull",
                "desc": "pull all instances of a value from an array",
                "field": kwargs.keys(),
                "value": kwargs.values()
                }
            }

    def increment(self, query:dict={}, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use increment method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        query.update({"_id": self.RECORD['_id']})

        self.PATCH(None, query, {"$inc": kwargs}, multi=True)
        self.reload()

        keys = list(kwargs.keys())
        values = [ kwargs[key] for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "increment",
                "desc": "increment the integer fields by the amount provided",
                "field": keys,
                "increment": values
                }
            }

    def update(self, query:dict={}, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use increment method on
 a non-existing record. Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        query.update({"_id": self.RECORD['_id']})

        keys = list(kwargs.keys())
        old_values = [ self.RECORD.get(key, None) for key in keys ]

        self.PATCH(None, query, {"$set": kwargs}, multi=True)
        self.reload()

        new_values = [ self.RECORD.get(key, None) for key in keys ]

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "action": "update",
                "desc": "replace existing field values with new values",
                "field": keys,
                "values_old": old_values,
                "values_new": new_values
                }
            }

    def patch(self, save=False, trigger=None, **kwargs):
        assert self.RECORD.get('_id'), """Cannot use patch method on
 a non-existing record.Use create method instead."""

        if "_id" in kwargs:
            kwargs.pop("_id")

        if self._DOC_MARSHMALLOW:
            _DOC_MARSHMALLOW().load(kwargs, partial=True)
            self.RECORD.update(kwargs)
        elif self.sample:
            assert all([ x in self.sample for x in kwargs.keys()])
            self.RECORD.update(kwargs)
        else:
            self.RECORD.update(kwargs)
        if save:
            self.save(trigger=trigger)

        return {
            "data": self._p_r(self.RECORD),
            "details": {
                "processed": True,
                "diff": kwargs
                }
            }

    def save(self, trigger=None):
        _id = None

        if self._DOC_DEFAULTS:
            for key, value in self._DOC_DEFAULTS.items():
                if not self.RECORD.get(key):
                    self.RECORD[key] = value

        if self.RECORD.get('_id'):
            _id = self.RECORD.pop("_id")
        try:
            if self._DOC_MARSHMALLOW:
                self._DOC_MARSHMALLOW().load(self.RECORD)
            else:
                validate(self.RECORD, self.schema)
        except:
            raise
        else:
            if _id:
                self.PATCH(None, _id, {"$set": self.RECORD})
                self.RECORD['_id'] = _id
            else:
                result = self.PUT(None, self.RECORD)
                self.RECORD['_id'] = result.inserted_id
            if trigger:
                trigger()

        return {
            "data": self._p_r(self.RECORD),
            "details": {}
            }

    def close(self):
        # TODO: clean closing logic
        self.load()

        return {
            "data": self.RECORD,
            "details": {}
            }

    def add_relation(self, key, *args, **kwargs):
        self.RECORD[key] = self._related_record(*args, **kwargs)

        return {
            "data": self._p_r(self.RECORD),
            "details": {"key": key, "relation": self.RECORD[key]}
            }

    def add_timestamp(self, key, value=None, relation={}):
        self.RECORD[key] = self._timestamp(value)

        if relation:
            related_key = relation.pop("key", self._guess_corresponding_fieldname(_type="datetime", related_field=key))
            self.add_relation(related_key, **relation)
        return {
            "data": self._p_r(self.RECORD),
            "details": {"key": key, "value": self.RECORD[key]}
            }

    def add_object(self, field, object_name=None, key=None, value=None, **kwargs):
        if not object_name:
            if not key:
                self.RECORD[field] = {}
            else:
                self.RECORD[field] = {key: value if value else self.GENERATE_ID()}
        else:
            template_path = os_path.join(TEMPLATE_PATH, object_name + ".json")
            assert os.path.exists(template_path), "path does not exist"
            with open(template_path) as file1:
                self.RECORD[field] = json_load(file1.read())
            if key:
                self.RECORD[field][key] = value if value else self.GENERATE_ID()

        return {
            "data": self._p_r(self.RECORD),
            "details": {field: self.RECORD[field], "state": "unsaved"}
            }


def get_client() -> SyncIOClient:
    """returns SyncIOClient class"""
    return SyncIOClient

def get_doc() -> SyncIODoc:
    """returns SyncIODoc class"""
    return SyncIODoc



#  vars.py
#
#  Copyright 2020 Anthony "antcer1213" Cervantes <anthony.cervantes@cerver.info>
#
#  MIT License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
import typing
from enum import Enum
import datetime
from bson.objectid import ObjectId
from pymongo import MongoClient
from dateutil.parser import parse as dateparse

# INFO: Objects
try:
    from pydantic import BaseConfig, BaseModel
    MODEL = typing.NewType("DataModel", BaseModel)
except:
    from dataclasses import dataclass
    MODEL = typing.NewType("DataModel", dataclass)

# INFO: Custom Types
MONGODB_URI = typing.NewType("MongoDB URI", str)
YAML = typing.NewType("YAML Document", str)
JSON = typing.NewType("JSON Document", str)
ENUM = typing.NewType("Enum", Enum)

# INFO: Static
ASCENDING = 1
DESCENDING = -1

# INFO: Fields
OBJECT_ID = OBJ_ID = ObjectId
DOC_ID = typing.NewType("Document ID", OBJ_ID)
DETAILS = typing.NewType("Meta Details", dict)

class StringEnum(str, Enum): pass
class IntEnum(int, Enum): pass

# NOTE: defaults to recommended fields; overwrite depending on your schema, use utils.generate_enum
PAGINATION_SORT_FIELDS = Enum(value="Pagination Sort Fields", names=[(item, item) for item in ("_id", "created_datetime", "updated_datetime")])


class ObjectIdStr(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, ObjectId):
            raise ValueError("Not a valid ObjectId")
        return str(v)


def str2bool(v):
    return str(v).lower() in ("yes", "true", "t", "1")


def str2datetime(v):
    if isinstance(v, (datetime.date, datetime,datetime)):
        return v
    else:
        return dateparse(v)

TYPES = {}
TYPES["str"] = str
TYPES["float"] = float
TYPES["int"] = int
TYPES["abs"] = abs
TYPES["dict"] = dict
TYPES["oid"] = ObjectId
TYPES["bool"] = ObjectId
TYPES["date"] = TYPES["datetime"] = str2datetime
TYPES["bool"] = str2bool

SCHEMA_TYPES = {}
SCHEMA_TYPES["str"] = SCHEMA_TYPES["string"] = SCHEMA_TYPES["text"] = "str"
SCHEMA_TYPES["number"] = SCHEMA_TYPES["num"] = SCHEMA_TYPES["decimal"] = SCHEMA_TYPES["float"] = "float"
SCHEMA_TYPES["int"] = SCHEMA_TYPES["integer"] = "int"
SCHEMA_TYPES["absolute"] = SCHEMA_TYPES["abs"] = "abs"
SCHEMA_TYPES["object"] = SCHEMA_TYPES["dict"] = SCHEMA_TYPES["obj"] = "dict"
SCHEMA_TYPES["oid"] = SCHEMA_TYPES["objectid"] = SCHEMA_TYPES["object_id"] = "oid"
SCHEMA_TYPES["date"] = "date"
SCHEMA_TYPES["datetime"] = "datetime"
SCHEMA_TYPES["bool"] = SCHEMA_TYPES["boolean"] = "bool"





import setuptools

with open("README.md", "r") as _file:
    long_description = _file.read()

with open("requirements.txt", "r") as _file:
    requirements = _file.readlines()[:-1]

setuptools.setup(
    name="cervmongo",
    version="0.1.20",
    author="Anthony 'antcer1213' Cervantes",
    author_email="anthony.cervantes@cerver.info",
    description="An (even) higher-level MongoDB client",
    keywords="motor pymongo mongodb database nonsql db mongo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/antcer1213/cervmongo",
    project_urls={
        "Bug Tracker": "https://github.com/antcer1213/cervmongo/issues",
        "Documentation": "https://cerver.info/packages/cervmongo/",
        "Source Code": "https://cerver.info/packages/cervmongo/_modules/",
        "Repository": "https://github.com/antcer1213/cervmongo",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=requirements,
    extras_require={
        "settings": ["python-dotenv"],
        "aio": ["motor"],
        "extra": ["pydantic", "marshmallow", "python-magic"],
        "all": ["python-dotenv", "motor", "pydantic", "marshmallow", "python-magic"],
    },
)
